package edu.georgetown.gucs.dictionary;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Random;
import java.util.HashMap;
import java.util.Vector;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.IllegalDataException;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import edu.georgetown.gucs.tokenizers.TokenizerList;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Token;


/**
 * Creates a list of unique tokens extracted from a collection of documents that can be trimmed by removing tokens based
 * on various different attributes; used for creating fingerprints of documents that are based on words that appear in a
 * document collection
 *
 * @author Clay Shields
 * @author Lindsay Neubauer
 */
public class Dictionary implements Serializable, Iterable<Entry<Token, DictionaryEntry>> {

    private static final long serialVersionUID = 8334786910169166049L;
    private boolean trimmed;
    private int totalDocuments;
    private int maxThread;
    private int nextPosition;
    private double maxIDFfound;
    private double trimmed_max = 1.0;
    private double trimmed_min = 0.0;
    private String creatingProgram;
    private String creation;
    private String dictionaryFileName;
    private String dictionaryName;
    private String directory;
    private String GUID;
    private String systemID;
    private Map<Token, DictionaryEntry> dictionary = new HashMap<Token, DictionaryEntry>();
    private TokenizerList tokenizers;
    private static final Logger logger = Logger.getLogger(Dictionary.class.getName());

    /**
     * Constructor that sets default values for this dictionary
     */
    public Dictionary() {
        this.dictionaryName = "";
        this.nextPosition = 0;
        this.totalDocuments = 0;
        boolean trained = false;
        this.maxIDFfound = 0.0;
        setCreatingProgram();
        setSystemID();
        this.GUID = new BigInteger(128, new Random()).toString();
        this.maxThread = Runtime.getRuntime().availableProcessors();
        this.trimmed = false;
        logger.log(Level.INFO, "[Constructor] Guid=", this.getGUID());
        logger.log(Level.INFO, "[Constructor] Max Thread=", this.maxThread);
        logger.log(Level.INFO, "[Constructor] trimmed=", this.trimmed);
    }

    /**
     * Constructor that loads this dictionary from a file
     *
     * @param dictionaryFilename the filename containing a dictionary
     */
    public Dictionary(String dictionaryFilename) {
        logger.log(Level.INFO,"Filename", dictionaryFilename);
        this.loadDictionaryXML(dictionaryFilename);
    }

    /**
     * Loads a previously serialized XML dictionary object into this dictionary
     *
     * @param filename the string filename of the XML dictionary file to load
     */
    @SuppressWarnings("unchecked")
    private void loadDictionaryXML(String filename) {
        this.dictionaryFileName = filename;

        try {
            File xml = new File(this.dictionaryFileName);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);
            Element dict = doc.getRootElement();

            this.dictionaryName = dict.getAttributeValue("name");


            this.GUID = dict.getAttributeValue("GUID");
            logger.log(Level.INFO, "GUID: " + this.getGUID());

            this.systemID = dict.getAttributeValue("systemID");
            logger.log(Level.INFO, "SystemID: " + this.systemID);


            this.creatingProgram = dict.getAttributeValue("creatingProgram");
            logger.log(Level.INFO, "CreatingProgram:", this.creatingProgram);

            this.creation = dict.getAttributeValue("creation");
            this.directory = dict.getAttributeValue("directory");
            logger.log(Level.INFO, "Directory:", this.directory);

            this.maxIDFfound = Double.valueOf(dict.getAttributeValue("maxIDF"));
            logger.log(Level.INFO, "MaxIDFfound:", this.maxIDFfound);

            this.totalDocuments = Integer.valueOf(dict.getAttributeValue("totalDocuments"));
            logger.log(Level.INFO, "TotalDocuments", this.totalDocuments);

            this.trimmed = Boolean.valueOf(dict.getAttributeValue("trimmed"));
            logger.log(Level.INFO, "Trimmed:", this.trimmed);

            if (this.trimmed) {

                this.trimmed_min = Double.valueOf(dict.getAttributeValue("trimmed_min"));
                logger.log(Level.INFO, "Trimmed_Min:", this.trimmed_min);

                this.trimmed_max = Double.valueOf(dict.getAttributeValue("trimmed_max"));
                logger.log(Level.INFO, "Trimmed_Max", this.trimmed_max);
            }

            this.nextPosition = Integer.valueOf(dict.getAttributeValue("nextPosition"));
            logger.log(Level.INFO, "next position:", this.nextPosition);

            Vector<String> tokenizerNames = new Vector<String>();
            for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
                String tokenizer = tokenizerElement.getAttributeValue("name");
                if (tokenizerElement.getAttribute("parameter") != null) {
                    tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
                }
                tokenizerNames.add(tokenizer.trim());
            }
            setTokenizers(tokenizerNames);
            logger.log(Level.INFO, "tokenizers", this.tokenizers);

            if (!this.dictionary.isEmpty()) {
                this.dictionary.clear();
            }

            //EVAN COME BACK
            for (Element tokenElement : (List<Element>) dict.getChildren("token")) {
                Token token = new Token(tokenElement.getAttributeValue("name"), Integer.valueOf(tokenElement
                        .getAttributeValue("start_byte")), Integer.valueOf(tokenElement.getAttributeValue("end_byte")));
                this.dictionary.put(
                        token,
                        new DictionaryEntry(token, Integer.valueOf(tokenElement.getAttributeValue("position")), Integer
                                .valueOf(tokenElement.getAttributeValue("frequency")), Integer.valueOf(tokenElement
                                .getAttributeValue("documentCount")), Double.valueOf(tokenElement.getAttributeValue("idf")))
                );
            }

        } catch (FileNotFoundException e) {
            ExceptionHandler.handleExceptionExit(e);
            logger.severe("File Not Found");
        } catch (UnsupportedEncodingException e) {
            ExceptionHandler.handleExceptionExit(e);
            logger.severe("Unsuported Encoding");
        } catch (JDOMException e) {
            ExceptionHandler.handleExceptionExit(e);
            logger.severe("JDOM Exception");
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
            logger.severe("IOException");
        }
    }

    /**
     * Reads an XML file that specifies which tokenizers to use for this dictionary
     *
     * @param config_file the string name of the XML configuration file containing this dictionary's tokenizers
     */
    @SuppressWarnings("unchecked")
    private void addTokenizers(String config_file) {

        logger.log(Level.INFO, "[Guid]", this.getGUID());
        logger.log(Level.INFO, "[Config File]", config_file);

        try {
            File xml = new File(config_file);
            SAXBuilder builder = new SAXBuilder();
            Document doc = builder.build(xml);

            Element dict = doc.getRootElement();

            Vector<String> tokenizerNames = new Vector<>();

            for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
                String tokenizer = tokenizerElement.getAttributeValue("name");

                logger.log(Level.INFO, "[tokenizer name]", tokenizer);
                if (tokenizerNames.size() == 0 && !tokenizer.contains("FileTokenizer")) {
                    ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: First tokenizer (" + tokenizer
                            + ") must read from a file."));
                }

                if (tokenizerElement.getAttribute("parameter") != null) {
                    tokenizer += " " + tokenizerElement.getAttributeValue("parameter");
                }
                tokenizerNames.add(tokenizer.trim());
            }
            setTokenizers(tokenizerNames);

            logger.log(Level.INFO, "[tokenizers]", this.tokenizers);
        } catch (FileNotFoundException e) {
            ExceptionHandler.handleExceptionExit(e);
        } catch (UnsupportedEncodingException e) {
            ExceptionHandler.handleExceptionExit(e);
        } catch (JDOMException e) {
            ExceptionHandler.handleExceptionExit(e);
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }

    /**
     * Creates a deep copy of the given dictionary object and stores it in this dictionary
     *
     * @param anotherDictionary the dictionary object to copy
     */
    public void clone(Dictionary anotherDictionary) {
        this.dictionary.clear();
        this.dictionary.putAll(anotherDictionary.dictionary);
        this.directory = anotherDictionary.directory;
        logger.log(Level.INFO, "[Directory]", this.directory);

        this.nextPosition = anotherDictionary.nextPosition;
        logger.log(Level.INFO, "[Next position]", this.nextPosition);

        this.trimmed = anotherDictionary.trimmed;
        logger.log(Level.INFO, "[Trimmed]", this.trimmed);
        if (this.trimmed) {
            this.trimmed_min = anotherDictionary.trimmed_min;
            logger.log(Level.INFO, "[Trimmed Min]", this.trimmed_min);

            this.trimmed_max = anotherDictionary.trimmed_max;
            logger.log(Level.INFO, "[Trimmed Max]", this.trimmed_max);
        }
        this.maxIDFfound = anotherDictionary.maxIDFfound;
        logger.log(Level.INFO, "[max IDF Found]", this.maxIDFfound);

        this.totalDocuments = anotherDictionary.totalDocuments;
        logger.log(Level.INFO, "[Total Documents]", this.totalDocuments);

        this.tokenizers = anotherDictionary.tokenizers;
        logger.log(Level.INFO, "[Tokenizers]", this.tokenizers);
    }

    /**
     * Determines if this dictionary contains a particular token
     *
     * @param token the string containing the token value
     * @return the <code>boolean</code> value of whether the token is present in this dictionary
     */
    public boolean containsToken(Token token) {
        return this.dictionary.containsKey(token);
    }

    /**
     * Creates this dictionary using the entire fileLister provided
     *
     * @param fileLister the fileLister containing the files to use for this dictionary
     */
    private void createFullDictionary(FileLister fileLister) {

        logger.log(Level.INFO, "[ENTER]");

        this.dictionary.clear();

        this.directory = fileLister.getDirectory().getAbsolutePath();

        ArrayList<DictionaryWorker> threads = new ArrayList<>(maxThread);
        File poison = new File(new BigInteger(128, new Random()).toString());

        for (int i = 0; i < maxThread; i++) {
            threads.add(i, new DictionaryWorker(this.tokenizers.getNames(), this, fileLister, poison));
            threads.get(i).setName("Dictionary Worker " + i);
            threads.get(i).setDaemon(true);
            threads.get(i).start();
        }

        // Wait for the file lister to finish loading files
        while (!fileLister.getLoaded()) {
            try {
                Thread.sleep(250);
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }
        }

        logger.log(Level.WARNING, "[FileLister complete]");

        // Now wait for the dictionary worker threads to finish
        boolean done = false;
        while (!done) {
            // Sleep a bit while the threads run
            try {
                Thread.sleep(100);
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }


            int waitingThreads = 0;
            for (int i = 0; i < threads.size(); i++) {
                if (threads.get(i).getState() == Thread.State.valueOf("WAITING")) {
                    waitingThreads++;
                }
            }

            // If so, then we are done loading.
            if (waitingThreads == threads.size()) {
                // Pass the poison to kill the worker threads
                fileLister.putPoison(poison);
                done = true;
            }

        }

        System.out.println("Workers completed");
        logger.log(Level.WARNING, "[Workers complete]");

        computeIDFs();
    }

    /**
     * Creates this dictionary using the optional count or percent in the given fileLister; if no count or percent is
     * specified then uses entire fileLister to create dictionary
     * <p/>
     * fileLister the fileLister containing the files to use for this dictionary, which needs to have completed loading
     */
    public void createPartialDictionary(List<File> dictionaryFiles, File directory) {

        if ((dictionaryFiles == null) || (dictionaryFiles.size() == 0)) {

            logger.log(Level.SEVERE, "Dying of empty or null fileList");
            throw new RuntimeException("Dictionary: createPartialDictionarary - nullFileLister received, exiting");
        }

        logger.log(Level.INFO, "[Enter]");
        this.dictionary.clear();

        this.directory = directory.getAbsolutePath();

        Iterator<File> dictionary_iterator = dictionaryFiles.iterator();

        LinkedBlockingQueue<File> sharedQueue = new LinkedBlockingQueue<>(128);

        File poison = new File("");

        // Create PartialDictionaryWorkerThreads to process the files
        ArrayList<PartialDictionaryWorker> threads = new ArrayList<>(maxThread);


        logger.log(Level.INFO, "[Starting Max Threads]", maxThread);
        for (int i = 0; i < maxThread; i++) {
            threads.add(i, new PartialDictionaryWorker(this.tokenizers.getNames(), this, sharedQueue, poison));
            threads.get(i).setName("PartialDictionaryWorker " + i);
            threads.get(i).setDaemon(true);
            threads.get(i).start();
        }

        // Put files on the queue for the workers to process
        while (dictionary_iterator.hasNext()) {
            try {
                File nextFile = dictionary_iterator.next();

                logger.log(Level.INFO, "[Adding Next File]", nextFile);
                sharedQueue.put(nextFile);
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }
        }

        // put the poison in the queue
        try {
            sharedQueue.put(poison);
        } catch (InterruptedException e) {
            ExceptionHandler.handleExceptionExit(e);
        }


        // wait a while for all threads to be waiting for more files, then interrupt them so they stop
        boolean done = false;
        while (!done) {

            try {

                logger.log(Level.INFO, "Waiting for Workers");
                Thread.sleep(100);


            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }


            // Check if the queue is empty and all threads in waiting state
            int waitingThreads = 0;
            for (int i = 0; i < threads.size(); i++) {
                if (threads.get(i).getState() == Thread.State.valueOf("TERMINATED")) {
                    waitingThreads++;
                }
            }
            logger.log(Level.INFO, "[Terminating workers]");
            if (waitingThreads == threads.size()) {
                done = true;
            }

        }

        computeIDFs();
    }

    /**
     * Allows for operations on the tokens in this dictionary
     *
     * @return the iterator over the tokens and their dictionaryEntry in this dictionary
     */
    @Override
    public synchronized Iterator<Entry<Token, DictionaryEntry>> iterator() {
        return this.dictionary.entrySet().iterator();
    }

    /**
     * Computes the Inverse Document Frequencies (IDFs) for each token in the dictionary and determines the largest IDF
     */
    private void computeIDFs() {
        int count = 0;
        Set<Entry<Token, DictionaryEntry>> theList = this.dictionary.entrySet();
        for (Entry<Token, DictionaryEntry> token : theList) {
            if (token == null) {
                logger.log(Level.SEVERE, "[Null token]");
                System.err.println("Null TOKEN!");
                throw new RuntimeException("Null Token");
            }
            double idf = token.getValue().computeIDF(this.totalDocuments);
            this.maxIDFfound = (idf > this.maxIDFfound) ? idf : this.maxIDFfound;
        }

        logger.log(Level.INFO, "[Max IDF Found]", this.maxIDFfound);
    }

    /**
     * Provides the program that created this dictionary
     *
     * @return the string name of the program that created this dictionary
     */
    public String getCreatingProgram() {
        return this.creatingProgram;
    }

    /**
     * Provides the maximum thread count to use for this dictionary
     *
     * @return the maximum thread count to use for this dictionary
     */
    public int getMaxThread() {
        return this.maxThread;
    }

    /**
     * Provides the original creation date of this dictionary
     *
     * @return the string date when this dictionary was created
     */
    public String getCreation() {
        return this.creation;
    }

    /**
     * Provides the filename of this dictionary, if this dictionary was loaded from or saved to a file
     *
     * @return the string dictionary filename
     */
    public String getDictionaryFilename() {
        return this.dictionaryFileName;
    }

    /**
     * Provides the name of this dictionary
     *
     * @return the string name of this dictionary
     */
    public String getDictionaryName() {
        return this.dictionaryName;
    }

    /**
     * Provides the unique identifier for this dictionary
     *
     * @return the string GUID for dictionary
     */
    public String getGUID() {
        return this.GUID;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMaxIDFFound() {
        return this.maxIDFfound;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMaxIDF() {
        return this.trimmed_max;
    }

    /**
     * Provides the largest IDF value in this dictionary
     *
     * @return the <code>double</code> value of the largest IDF in this dictionary
     */
    public double getMinIDF() {
        return this.trimmed_min;
    }

    /**
     * Provides the normalized inverse document frequency (IDF) of the given token in this dictionary. The normalized IDF
     * is the token's IDF divided by the highest IDF in this dictionary.
     *
     * @param token the string containing the token value
     * @return the <code>double</code> value of the token's normalized inverse document frequency (IDF); returns 0.0 if
     * the token is not present in this dictionary
     */
    public synchronized double getNormalizedIDF(Token token) {

        if (this.dictionary.containsKey(token)) {
            return getNormalizedIDF(this.dictionary.get(token));
        }

        return 0.0;
    }

    /**
     * Provides the normalized inverse document frequency (IDF) of the given dictionaryEntry. The normalized IDF is the
     * token's IDF divided by the highest IDF in this dictionary.
     *
     * @param entry the dictionaryEntry to calculate the normalized IDF
     * @return the <code>double</code> value of the dictionaryEntry's normalized inverse document frequency (IDF)
     */
    public double getNormalizedIDF(DictionaryEntry entry) {
        return entry.getIDF() / this.maxIDFfound;
    }

    /**
     * Provides the position of the given token in this dictionary
     *
     * @param token the string containing the token value
     * @return the position of the token; returns -1 if the token is not present in this dictionary
     */
    public int getPosition(Token token) {
        if (this.dictionary.containsKey(token)) {
            return this.dictionary.get(token).getPosition();
        }
        return -1;
    }

    /**
     * Provides the number of tokens in this dictionary
     *
     * @return the number of unique tokens in this dictionary
     */
    public int size() {
        return this.dictionary.size();

    }

    /**
     * Provides the directory used for this dictionary
     *
     * @return the string directory for which this dictionary was created
     */
    public String getSource() {
        return this.directory;
    }

    /**
     * Provides the identifier for the system that this dictionary was created on
     *
     * @return the string system ID for this dictionary
     */
    public String getSystemID() {
        return this.systemID;
    }

    /**
     * Provides the tokenizerList object used to create this dictionary
     *
     * @return the tokenizerList object used for this dictionary
     */
    public TokenizerList getTokenizers() {
        return this.tokenizers;
    }

    /**
     * Provides the names of the tokenizers
     *
     * @return the names of the tokenizers used for this dictionary
     */
    public List<String> getTokenizerNames() {
        return this.tokenizers.getNames();
    }


    /**
     * Provides a vector of the tokens in this dictionary
     *
     * @return the list of tokens in this dictionary
     */
    public synchronized List<Token> getTokens() {

        return Collections.synchronizedList(new Vector<Token>(this.dictionary.keySet()));

    }

    /**
     * Provides the number of documents processed for this dictionary
     *
     * @return the total number of documents used to create this dictionary
     */
    public int getTotalDocuments() {
        return this.totalDocuments;
    }

    /**
     * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs
     * dictionary to a file
     *
     * @param config_file the string name of the XML configuration file containing dictionary tokenizers
     * @param path        the string path to the directory used to create this dictionary
     * @param output_file the string name/path to write this dictionary to
     */
    public void makeDictionary(String config_file, String path, String output_file) {


        logger.log(Level.INFO, "Config", config_file);
        logger.log(Level.INFO, "Config", path);
        logger.log(Level.INFO, "Config", output_file);

        if (config_file == null || path == null) {
            logger.log(Level.SEVERE, "Null Config OR Path paramaters");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null parameter(s) given."));
        }

        addTokenizers(config_file);

        if (this.tokenizers.getNames() == null) {
            logger.log(Level.SEVERE, "Null TokenizerList");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null tokenizerList"));
        }
        if (this.tokenizers.getNames().size() < 1) {
            logger.log(Level.SEVERE, "Empty TokenizerList");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Empty tokenizerList"));
        }
        if (this.dictionary == null) {
            logger.log(Level.SEVERE, "Null Dictionary");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null dictionary"));
        }

        FileLister files = new FileLister(path, this.maxThread);

        files.loadQueue();

        createFullDictionary(files);

        if (output_file == null) {
            logger.log(Level.SEVERE, "Null output file");
            stdOutDictionaryXML();
        } else {
            logger.log(Level.SEVERE, "Standard out ");
            saveDictionaryXML(output_file);
        }
    }

    /**
     * Creates dictionary from the given directory; Reads an XML file that specifies which tokenizers to use and outputs a
     * trimmed dictionary to a file. Used by the GUI, not the command line version
     *
     * @param tokenizersNames the list of names of the dictionary tokenizers
     * @param path            the string path to the directory used to create this dictionary
     * @param min_idf         the <code>double</code> minimum normalized inverse document frequency (IDF) to trim dictionary
     * @param max_idf         the <code>double</code> maximum normalized inverse document frequency (IDF) to trim dictionary
     * @param output_file     the string name/path to write this dictionary to
     */
    public void makeDictionary(List<String> tokenizersNames, String path, double min_idf, double max_idf,
                               String output_file) {
        if (output_file == null || tokenizersNames == null || path == null) {
            logger.log(Level.SEVERE, "Null output file OR tokenizerName OR Path");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null parameter(s) given."));
        }

        setTokenizers(tokenizersNames);

        if (this.tokenizers.getNames() == null) {
            logger.log(Level.SEVERE, "Null TokenizerList");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null tokenizerList"));
        }
        if (this.tokenizers.getNames().size() < 1) {
            logger.log(Level.SEVERE, "Empty TokenizerList");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Empty tokenizerList"));
        }
        if (this.dictionary == null) {
            logger.log(Level.SEVERE, "Null Dictionary");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Null dictionary"));
        }

        FileLister files = new FileLister(path, this.maxThread);
        createFullDictionary(files);

        if (min_idf > 0 || max_idf < 1.0) {
            logger.log(Level.INFO, "Min IDF", min_idf);
            logger.log(Level.INFO, "Max IDF", max_idf);
            trimByIDF(min_idf, max_idf);
        }
        saveDictionaryXML(output_file);
    }

    /**
     * Merges the given Map of tokens to dictionaryEntry with this dictionary
     *
     * @param documentTokens the map containing the dictionary information to merge with this dictionary
     */
    protected synchronized void mergeDictionary(Map<Token, DictionaryEntry> documentTokens) {
        for (Entry<Token, DictionaryEntry> documentEntry : documentTokens.entrySet()) {
            if (this.dictionary.containsKey(documentEntry.getKey())) {
                // If so, increment the document count and frequency count
                this.dictionary.get(documentEntry.getKey()).incrementFrequencyCount(
                        documentEntry.getValue().getFrequencyCount());
                this.dictionary.get(documentEntry.getKey()).incrementDocumentCount();
            } else {
                // Otherwise, insert the DictionaryEntry into this dictionary
                documentEntry.getValue().setPosition(this.nextPosition);
                this.nextPosition++;
                this.dictionary.put(documentEntry.getKey(), documentEntry.getValue());
            }
        }
        this.totalDocuments++;
    }

    /**
     * Writes this dictionary as a serialized XML object.
     *
     * @param filename the string XML filename to write this dictionary to
     */
    void saveDictionaryXML(String filename) {
        logger.log(Level.INFO, "Max Threads", this.maxThread);

        this.dictionaryFileName = (new File(filename)).getAbsolutePath();
        this.dictionaryName = filename.substring(filename.lastIndexOf('/') + 1);
        this.creation = String.valueOf(new Date());
        DecimalFormat rounding = new DecimalFormat("#0.####");

        Element dict = new Element("dictionary");
        dict.setAttribute("name", this.dictionaryFileName);
        dict.setAttribute("GUID", this.GUID);
        dict.setAttribute("systemID", this.systemID);
        if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
            dict.setAttribute("creatingProgram", this.creatingProgram);
        if (getCreation() != null && getCreation().length() > 0)
            dict.setAttribute("creation", this.creation);
        if (getSource() != null && getSource().length() > 0)
            dict.setAttribute("directory", this.directory);
        dict.setAttribute("maxIDF", rounding.format((this.maxIDFfound)));
        dict.setAttribute("totalDocuments", String.valueOf(this.totalDocuments));
        dict.setAttribute("trimmed", String.valueOf(this.trimmed));
        if (this.trimmed) {
            dict.setAttribute("trimmed_min", String.valueOf(this.trimmed_min));
            dict.setAttribute("trimmed_max", String.valueOf(this.trimmed_max));
        }
        dict.setAttribute("nextPosition", String.valueOf(this.nextPosition));

        for (String tokenizer : this.tokenizers.getNames()) {
            String[] split = tokenizer.split("\\s");
            Element tokenizerElement = new Element("tokenizer");
            tokenizerElement.setAttribute("name", split[0]);
            StringBuffer parameters = new StringBuffer();
            if (split.length > 1) {
                parameters.append(split[1]);
                tokenizerElement.setAttribute("parameter", split[1]);
            }
            dict.addContent(tokenizerElement);
        }

        for (Entry<Token, DictionaryEntry> entry : this) {
            try {
                Element tokenElement = new Element("token");
                tokenElement.setAttribute("name", entry.getKey().name());
                tokenElement.setAttribute("start_byte", String.valueOf(entry.getKey().start_byte()));
                tokenElement.setAttribute("end_byte", String.valueOf(entry.getKey().end_byte()));
                tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
                tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
                tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
                tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
                tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
                dict.addContent(tokenElement);
            } catch (IllegalDataException exception) {
                logger.log(Level.SEVERE, "Dictionary org.jdom.IllegalDataException prohibits " +
                        "token from being added to dictionary:", entry.getKey());

                System.err.println("Dictionary.org.jdom.IllegalDataException prohibits token from " +
                        "being added to dictionary: " + entry.getKey());
            }
        }

        try {
            Document doc = new Document(dict);
            FileOutputStream dictionaryOut = new FileOutputStream(this.dictionaryFileName);
            Format form = Format.getPrettyFormat();
            form.setEncoding(Global.fileEncoding);
            XMLOutputter out = new XMLOutputter(form);
            out.output(doc, dictionaryOut);

            System.out.println("Writing dictionary to " + this.dictionaryFileName);
            logger.log(Level.INFO, "Writing dictionary to:", this.dictionaryFileName);
            dictionaryOut.close();
        } catch (FileNotFoundException e) {
            ExceptionHandler.handleExceptionExit(e);
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }

    /**
     * Writes dictionary to standard output
     */
    void stdOutDictionaryXML() {
        logger.log(Level.INFO, "Save Ductuibary XML: MaxThreads", this.maxThread);

        this.creation = String.valueOf(new Date());
        DecimalFormat rounding = new DecimalFormat("#0.####");

        Element dict = new Element("dictionary");
        //  dict.setAttribute("name", this.dictionaryFileName);
        dict.setAttribute("GUID", this.GUID);
        dict.setAttribute("systemID", this.systemID);
        if (getCreatingProgram() != null && getCreatingProgram().length() > 0)
            dict.setAttribute("creatingProgram", this.creatingProgram);
        if (getCreation() != null && getCreation().length() > 0)
            dict.setAttribute("creation", this.creation);
        if (getSource() != null && getSource().length() > 0)
            dict.setAttribute("directory", this.directory);
        dict.setAttribute("maxIDF", rounding.format((this.maxIDFfound)));
        dict.setAttribute("totalDocuments", String.valueOf(this.totalDocuments));
        dict.setAttribute("trimmed", String.valueOf(this.trimmed));
        if (this.trimmed) {
            dict.setAttribute("trimmed_min", String.valueOf(this.trimmed_min));
            dict.setAttribute("trimmed_max", String.valueOf(this.trimmed_max));
        }
        dict.setAttribute("nextPosition", String.valueOf(this.nextPosition));

        for (String tokenizer : this.tokenizers.getNames()) {
            String[] split = tokenizer.split("\\s");
            Element tokenizerElement = new Element("tokenizer");
            tokenizerElement.setAttribute("name", split[0]);
            StringBuffer parameters = new StringBuffer();
            if (split.length > 1) {
                parameters.append(split[1]);
                tokenizerElement.setAttribute("parameter", split[1]);
            }
            dict.addContent(tokenizerElement);
        }

        for (Entry<Token, DictionaryEntry> entry : this) {
            try {
                Element tokenElement = new Element("token");
                tokenElement.setAttribute("name", entry.getKey().name());
                tokenElement.setAttribute("start_byte", String.valueOf(entry.getKey().start_byte()));
                tokenElement.setAttribute("end_byte", String.valueOf(entry.getKey().end_byte()));
                tokenElement.setAttribute("position", String.valueOf(entry.getValue().getPosition()));
                tokenElement.setAttribute("frequency", String.valueOf(entry.getValue().getFrequencyCount()));
                tokenElement.setAttribute("documentCount", String.valueOf(entry.getValue().getDocumentCount()));
                tokenElement.setAttribute("idf", rounding.format((entry.getValue().getIDF())));
                tokenElement.setAttribute("idfNorm", rounding.format((getNormalizedIDF(entry.getValue()))));
                dict.addContent(tokenElement);
            } catch (IllegalDataException exception) {
               System.err.println("Dictionary: org.jdom.IllegalDataException prohibits token from " +
                       "being added to dictionary: " + entry.getKey());
                logger.log(Level.SEVERE, "org.jdom.IllegalDataException prohibits token" +
                        " from being added to dictionary: ", entry.getKey());
            }
        }

        try {
            Document doc = new Document(dict);
            Format form = Format.getPrettyFormat();
            form.setEncoding(Global.fileEncoding);
            XMLOutputter printer = new XMLOutputter(form);
            printer.output(doc, System.out);
            System.out.println("Writing dictionary to standard out");
            logger.log(Level.INFO, "Writing Dictionary to Standard Out [STD]");
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }

    /**
     * Determines the name of the program that created this dictionary
     */
    private void setCreatingProgram() {
        logger.info("Start");
        this.creatingProgram = "";
        Runtime runtime = Runtime.getRuntime();
        String cmds[] = {"jps", "-m"};
        Process proc;
        try {
            proc = runtime.exec(cmds);
            BufferedReader bufferedreader = new BufferedReader(new InputStreamReader(proc.getInputStream(), Global.fileEncoding));

            String line;
            while ((line = bufferedreader.readLine()) != null) {
                for (String output : line.split("\\s")) {
                    if (output.contains(".jar")) {
                        this.creatingProgram = output;
                        logger.log(Level.INFO, "Creating Program", this.creatingProgram);
                    }
                }
            }
            bufferedreader.close();
            proc.getInputStream().close();
            proc.getOutputStream().flush();
            proc.getOutputStream().close();
            proc.getErrorStream().close();

        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
    }

    /**
     * Sets the name of the program that created this dictionary
     *
     * @param creatingProgram the string program that created this dictionary
     */
    public void setCreatingProgram(String creatingProgram) {
        this.creatingProgram = creatingProgram;
        logger.log(Level.INFO, "Creating Program", creatingProgram);
    }

    /**
     * Sets the system ID of this dictionary
     */
    private void setSystemID() {
        try {
            this.systemID = java.net.InetAddress.getLocalHost().getCanonicalHostName();
        } catch (UnknownHostException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
        logger.log(Level.INFO, "System ID", this.systemID);
    }

    /**
     * Loads the tokenizers to use for this dictionary from a TokenizerList object. Only works if no tokenizers have
     * already been set.
     *
     * @param tokenizerList the TokenizerList object containing the tokenizers to use for this dictionary
     */
    public void setTokenizers(TokenizerList tokenizerList) {
        if (this.tokenizers == null) {
            this.tokenizers = tokenizerList;
        } else {
            logger.log(Level.SEVERE, "Dictionary Cannot Reset Tokenizer List in a dictionary");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Cannot reset tokenizer list in a dictionary"));
        }
        logger.log(Level.INFO, "Tokenizers", this.tokenizers);
    }

    /**
     * Sets the tokenizers used by this dictionary; Only works if no tokenizers have already been set or if the given list
     * contains the same tokenizers as those that have already been set
     *
     * @param tokenizerVec the list of tokenizer names to use to create this dictionary
     */
    public void setTokenizers(List<String> tokenizerVec) {
        logger.log(Level.INFO, "tokenizers", tokenizerVec);
        if (this.tokenizers == null) {
            this.tokenizers = new TokenizerList(tokenizerVec);
        } else if (this.tokenizers.getNames().size() != tokenizerVec.size()
                || !tokenizerVec.containsAll(this.tokenizers.getNames())) {
            logger.log(Level.SEVERE, "Dictionary cannot reset tokenizers in a dictionary");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Cannot reset tokenizers in a dictionary"));
        }
        logger.log(Level.INFO, "tokenizers", this.tokenizers);
      }

    /**
     * Prints this dictionary statistics, including:
     * <ul>
     * <li>Name</li>
     * <li>Number of documents</li>
     * <li>Number of tokens</li>
     * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
     * <li>Minimum and maximum IDF</li>
     * <li>List of tokenizers</li>
     * </ul>
     */
    void showStatistics() {
        System.out.println("Name: " + this.dictionaryFileName);
        System.out.println("# Docs: " + this.totalDocuments);
        System.out.println("# Tokens: " + this.dictionary.size());
        if (this.trimmed) {
            System.out
                    .println("Trimmed: " + this.trimmed + " (min: " + this.trimmed_min + ", max: " + this.trimmed_max + ")");
        } else {
            System.out.println("Trimmed: " + this.trimmed);
        }
        System.out.println("Max IDF: " + this.maxIDFfound);
        System.out.println("Tokenizers: " + this.tokenizers);
    }

    /**
     * Provides this dictionary statistics, including:
     * <ul>
     * <li>Name</li>
     * <li>Number of documents</li>
     * <li>Number of tokens</li>
     * <li>If this dictionary has been trimmed (including the IDF range, if trimmed)</li>
     * <li>Minimum and maximum IDF</li>
     * <li>List of tokenizers</li>
     * </ul>
     *
     * @return the string containing this dictionary's statistics
     */
    public String getStatistics() {
        StringBuffer statisticsBuffer = new StringBuffer();
        statisticsBuffer.append(String.format("%s%n", "Name: " + this.dictionaryFileName));
        statisticsBuffer.append(String.format("%s%n", "# Docs: " + this.totalDocuments));
        statisticsBuffer.append(String.format("%s%n", "# Tokens: " + this.dictionary.size()));

        if (this.trimmed) {
            statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed + " (min: " + this.trimmed_min
                    + ", max: " + this.trimmed_max + ")"));
        } else {
            statisticsBuffer.append(String.format("%s%n", "Trimmed: " + this.trimmed));
        }
        statisticsBuffer.append(String.format("%s%n", "Max IDF: " + this.maxIDFfound));
        statisticsBuffer.append(String.format("%s%n", "Tokenizers: " + this.tokenizers));
        return statisticsBuffer.toString();
    }

    /**
     * Prints each token in this dictionary with it's frequency, IDF, and normalized IDF
     */

    void showTokens() {
        System.out.format("%-25s %-10s %-10s %-10s%n", "Token", "Freq", "IDF", "Norm. IDF");
        System.out.println("----------------------------------------------------------");
        for (Entry<Token, DictionaryEntry> entry : this) {
            System.out.format("%-25s %-10s %-10.4f %-10.4f%n", entry.getKey(), entry.getValue().getFrequencyCount(), entry
                    .getValue().getIDF(), getNormalizedIDF(entry.getValue()));
        }
    }

    /**
     * Turns on threading for creating this dictionary. This should only be used for stand-alone dictionary creation (not
     * inside a threaded application).
     *
     * @param numThreads the number of threads to use to create this dictionary
     */
    public void threadingOn(int numThreads) {
        this.maxThread = numThreads;
        logger.log(Level.INFO, "Max Thread", this.maxThread);
    }

    /**
     * Trims this dictionary by removing any token that is outside a range of normalized IDFs. Once trimmed, no new tokens
     * can be added to this dictionary.
     *
     * @param min the <code>double</code> minimum normalized IDF to keep; any token with a lower IDF will be discarded.
     * @param max the <code>double</code> maximum normalized IDF to keep; any token with a higher IDF will be discarded.
     */
    public void trimByIDF(double min, double max) {
        logger.log(Level.INFO, "start");

        if (min >= max) {
            logger.log(Level.SEVERE, "Invalid IDF Range");
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: Invalid IDF range:  min_idf=" + min
                    + " max_idf=" + max));
        }

        // Trims this dictionary to contain only tokens within the specified range of normalized IDF values
        if (this.trimmed) {
            System.err.println("Dictionary trimming an already trimmed dictionary might be a bad idea");
            logger.log(Level.INFO, "Trimming an already trimmed Dictionary");
        }

        this.trimmed = true;
        this.trimmed_min = min;
        this.trimmed_max = max;
        logger.log(Level.INFO, "trimmed", this.trimmed);
        logger.log(Level.INFO, "trimmed min", this.trimmed_min);
        logger.log(Level.INFO, "trimmed max", this.trimmed_max);
        int newposition = 0;
        HashMap<Token, DictionaryEntry> newdict = new HashMap<Token, DictionaryEntry>();

        Iterator<Entry<Token, DictionaryEntry>> iterator = this.dictionary.entrySet().iterator();

        while (iterator.hasNext()) {
            Entry<Token, DictionaryEntry> entry = iterator.next();
            if (!(getNormalizedIDF(entry.getValue()) < min || getNormalizedIDF(entry.getValue()) > max)) {
                entry.getValue().setPosition(newposition);
                newposition++;
                newdict.put(entry.getKey(), entry.getValue());
            }

        }
        this.dictionary = newdict;
        logger.info("end");
        }

    /**
     * Creates a dictionary from the specified directory using the provided tokenizers
     *
     * @param args array of string command line arguments<br>
     *             <code>args[0]</code> the filename of the tokenizer configuration file<br>
     *             <code>args[1]</code> the directory to create the dictionary from<br>
     *             <code>args[2]</code> the filename to write the dictionary to
     */
    public static void main(String args[]) {
        String output_file = null;
        String directory = null;
        String config_file = null;

        try {
            config_file = args[0];
            directory = args[1];
            output_file = args[2];
        } catch (ArrayIndexOutOfBoundsException e) {
            System.err
                    .println("Usage: java CreateDictionary <tokenizers config file> <directory> <output dictionary filename> ");
            System.err.println("Recursively creates a dictionary from the specified directory using the provided tokenizers");
            System.exit(-1);
        }

        try {
            Dictionary dict = new Dictionary();
            dict.makeDictionary(config_file, directory, output_file);
        } catch (RuntimeException exception) {
            ExceptionHandler.handleExceptionExit(exception);
        }
    }
}
