package edu.georgetown.gucs.experiment;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.matcher.FingerprintMatcher;
import edu.georgetown.gucs.utility.*;

import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Logger;

/**
 * @author Evan B.
 * @since 6/17/14
 */
public class Trial extends Thread {

    private TrialParameters poison;
    private LinkedBlockingQueue<TrialParameters> parameterQueue;
    private Experiment parent;
    private TrialParameters base_parameters;

    private int experimentID;
    private int minScore;

    private Random rnd;

    private String dataset;
    private String matcherName;
    private String fingerprinterName;
    private String mangler;
    private String splitter;

    private Fingerprinter fingerprinter;

    private FileLister fileSource;

    private double dictionarySampleSize;
    private double sampleSize;

    private List<String> tokenizerNames;
    private Pair<Double, Double> idfRange;


    private ExperimentResults results;

    private static final Logger logger = Logger.getLogger(Trial.class.getName());

    /**
     * @param theQueue  all of the parameters
     * @param thePoison indicates when the queue is empty
     * @param parent    the class to pass results
     */
    Trial(LinkedBlockingQueue<TrialParameters> theQueue,
          TrialParameters thePoison,
          Experiment parent) {

        this.parameterQueue = theQueue;
        this.poison = thePoison;
        this.parent = parent;
    }

    /**
     * Runs the trial
     */
    @Override
    public void run() {

        String threadName = Thread.currentThread().getName();

        boolean done = false;
        logger.info("[entering param while loop [done=" + done + "]");

        while (!done) {


            // Grab a set of parameters from the queue
            try {
                base_parameters = parameterQueue.take();
                //System.out.println("[Parameter queue size] " + parameterQueue.size());
                logger.info("taking base_params [queue size] " + parameterQueue.size());

            } catch (InterruptedException e) {
                logger.info("interrupted exception baseParams failed to take");
                ExceptionHandler.handleExceptionExit(e);
            }
            //  System.out.println("[paramQueue size] " + parameterQueue.size());

            //initialize object instance fields
            unpackParams();

            logger.info("finished unpacking params:");

            /**
             * Debug Code
             * printParams();
             * if(parameterQueue.size() > 0 ) {
             * System.out.println(poison.toString());
             * System.out.println(base_parameters.toString() + "-------------------------\n\n");
             * }
             */

            // Quit if we get the poison, and pass it on and exit
            if (base_parameters == poison) {
                try {
                    logger.info("putting poison on queue and quitting");
                    System.out.println("[Terminating] putting poison on queue");
                    parameterQueue.put(poison);
                } catch (InterruptedException e) {
                    ExceptionHandler.handleExceptionExit(e);
                    logger.info("interruptedException");
                }
                logger.info("Throwing a linkedblockingQueue exception");
                System.out.println("Return from poison [kill thread] ");
                return;
            }

            // System.out.println("Initializing Results");
            //initialize experimentResults
            results = new ExperimentResults(base_parameters);
            logger.info("[initializing results ");

            results.setThreadName(threadName);
            logger.info("setting threadName " + threadName);

            if (this.base_parameters == null) {
                logger.info("base_params null [ERROR]");
                System.err.println("Trial " + threadName + "Thread called with null parameters");
                throw new RuntimeException("Trial Thread called with null parameters");
            }

            //  System.out.println("Trial: Getting sublists from fileLister: "
            //        + fileSource.getDirectory());

            Pair<List<File>, List<File>> dictAndSampleLists =
                    fileSource.getSublists(dictionarySampleSize, sampleSize);
            logger.info("DictAndSampleLists A " + dictAndSampleLists.getA().size());
            logger.info("DictAndSampleLists B " + dictAndSampleLists.getB().size());

            List<File> dictionaryList = dictAndSampleLists.getA();
            List<File> sampleList = dictAndSampleLists.getB();

            logger.info("Starting to create Dictionary");
            int dictThreads = 4;
            Dictionary dictionary = new Dictionary();
            dictionary.threadingOn(dictThreads);

            dictionary.setTokenizers(tokenizerNames);
            dictionary.createPartialDictionary(dictionaryList, fileSource.getDirectory());
            logger.info("[FileSource Directory] " + fileSource.getDirectory());

            if ((idfRange.getA() > 0) || (idfRange.getB() < 1)) {
                dictionary.trimByIDF(idfRange.getA(), idfRange.getB());
                logger.info("idfRange in bounds");
            }

            int dictionarySize = dictionary.size();
            logger.info("dictionary size " + dictionary.size());

            results.setDictionarySize(dictionarySize);


            fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);

            FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);
            matcher.setMinimumScore(minScore);

            //here is the problem
            logger.info("entering computeFingerprints");
            computeFingerprints(dictionary, sampleList);

            logger.info("get results [FINISHED COMPUTE FINGERPRINTS");

            parent.getResults(results);
        }
    }


    private void unpackParams() {

        matcherName = base_parameters.getMatcherName();

        dataset = base_parameters.getDataset();
        sampleSize = base_parameters.getSampleParameter();
        idfRange = base_parameters.getIDFRange();
        minScore = base_parameters.getMinScore();
        dictionarySampleSize = base_parameters.getDictionarySampleParameter();

        fingerprinterName = base_parameters.getFingerprinterName();
        fingerprinter = null;

        fileSource = base_parameters.getFileSource();
        tokenizerNames = base_parameters.getTokenizerNames();

        mangler = base_parameters.getMangler();
        splitter = base_parameters.getSplitter();

        rnd = base_parameters.getRandom();
    }


    /**
     * If the mangler is null --> compute non fingerprint
     * if the mangler is not null --> compute fingerprint with mangler
     * store unmangled fingerprint
     * compare mangled fingerprint of the same dataSets;
     * --> score
     */

    //mangle and unmangled fingerprints
    private void computeFingerprints(Dictionary dictionary, List<File> fileSet) {

        FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);
        matcher.setMinimumScore(minScore);


        ArrayList<Integer> scores = new ArrayList<>();
        for (File file : fileSet) {

            List<Fingerprint> mangled;
            List<Fingerprint> unmangled;

            //Current file being edited
            String filePath = file.getAbsolutePath();

            try {

                fingerprinter.setDictionary(dictionary);

                if (!mangler.contains("null")) {
                    fingerprinter.setMangler(mangler, dictionary);
                    fingerprinter.setManglerRNG(rnd);
                    fingerprinter.setSplitter(splitter);
                }
                mangled = fingerprinter.computeFingerprint(filePath);

                dictionary.getTokenizers().disableMangler();
                fingerprinter.setDictionary(dictionary);
                unmangled = fingerprinter.computeFingerprint(filePath);


                if (mangled.size() <= 0) {
                    //debug output
                    System.out.println("File: " + filePath);
                    System.out.println("[mangler] " + mangler);
                    System.out.println("[Fingerprinter] " + fingerprinterName);
                    System.out.println("[unmangled size] " + unmangled.size());
                    System.out.println("[mangled size]  " + mangled.size());
                    System.out.println("[tokenizer Names] " + tokenizerNames);
                }

                //ensure fingerprints were computed
                try {

                    int score = matcher.getScore(unmangled.get(0).getFingerprint(),
                            mangled.get(0).getFingerprint());

                    //add scores to later average
                    scores.add(score);

                    /*
                    System.out.println("unamngledSize " + unmangled.get(0).getTokenList().size());
                    System.out.println("mangledSize " + mangled.get(0).getTokenList().size());
                    System.out.println("Matcher " + matcher.getScore(unmangled.get(0).getFingerprint(),
                            mangled.get(0).getFingerprint()));
                    */

                } catch (IndexOutOfBoundsException e) {
                    // e.printStackTrace();
                    logger.info("array out of bounds index with mangler");
                }

            } catch (IOException e) {
                e.printStackTrace();

            }

        }

        //average scores
        int count = 0;
        for (Integer i : scores) {
            count += i;
        }

        results.setMatcherScore((count / scores.size()));

    }
}
