package edu.georgetown.gucs.matcher;

import java.util.SortedMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.List;
import java.util.Vector;
import java.util.Collections;
import java.util.ArrayList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Pair;
import edu.georgetown.gucs.utility.Token;

/**
 * This is a class that allows threaded parallel comparison of fingerprints.
 *
 * @author Clay Shields
 */

public class BulkMatcher {

    private final String matcherName;
    private final int minScore;
    private final int maxThread;

    // internal results is a vector from 0 to 100, one for each possible score
    // Each entry is a list of matching file pairs
    private final List<List<Pair<Token, Token>>> internalResults;
    private final SortedMap<Integer, List<String>> results;
    private final LinkedBlockingQueue<Pair<Integer, Integer>> matchingQueue;
    private final AtomicInteger finishedThreads;

    private static final Logger logger = Logger.getLogger(BulkMatcher.class.getName());

    /**
     * A constructor for when you want to limit the number of threads that are run
     * @param results a map to insert the results in
     * @param maxThread limit on number of threads to be used
     */
    BulkMatcher(String matcherType, SortedMap<Integer, List<String>> results, int maxThread) {


        this.matcherName = matcherType;
        this.results = results;
        this.minScore = 0;
        this.maxThread = maxThread;
        this.matchingQueue = new LinkedBlockingQueue<>();
        this.internalResults = Collections.synchronizedList(new ArrayList<>(Global.MAX_SCORE));

        for (int i = 0; i <= Global.MAX_SCORE; i++) {
            this.internalResults.add(i, Collections.synchronizedList(new ArrayList<>()));
        }

        finishedThreads = new AtomicInteger(0);

        logger.info("BulkMatcher " + " constructor" +  " Matcher: "
                + this.matcherName + " min score: " + this.minScore + " maxThread: " + maxThread);

    }

    /**
     * A method to do n^2 comparisons among all fingeprints present
     *
     * @param fingerprints a list of the fingerprint/files pairs to compare
     */
    public void compareAllFingerprints(List<Fingerprint> fingerprints) {

        logger.log(Level.INFO, "Number of fingerprints", fingerprints.size());

        Pair<Integer, Integer> poison = new Pair(-1, new Integer(-1));
        ArrayList<BulkMatcherWorker> threads = new ArrayList<>(maxThread);

        logger.log(Level.INFO, "Starting threads", maxThread);

        for (int i = 0; i < maxThread; i++) {
            threads.add(i, new BulkMatcherWorker(this.matcherName, this.minScore,
                    matchingQueue, fingerprints, this.internalResults, poison, finishedThreads));

            threads.get(i).setName("Bulk Matcher Worker " + i);
            threads.get(i).setDaemon(true);
            threads.get(i).start();
        }


        // Fill the queue with columns for each thread
        for (Integer i = 0; i < fingerprints.size() - 2; i++) {
            try {
                matchingQueue.put(new Pair(i, fingerprints.size()));
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }
        }


        // Now wait for the threads to finish processing
        boolean done = false;
        while (!done) {
            // Sleep a bit while the threads run
            try {
                logger.log(Level.INFO, "Awaiting thread termination");
                Thread.sleep(100);
            } catch (InterruptedException e) {
                ExceptionHandler.handleExceptionExit(e);
            }

            // If there are elements left, we are not done
            if (matchingQueue.peek() == null) {
                // If the elements are gone, wait to see how many threads are running
                int waitingThreads = 0;
                for (int i = 0; i < threads.size(); i++) {
                    if (threads.get(i).getState() == Thread.State.valueOf("WAITING")) {
                        waitingThreads++;
                    }
                }
                // If so, then we are done loading and matching.
                if (waitingThreads == threads.size()) {
                    logger.info("terminating threads");
                    // Pass the poison to kill the worker threads
                    try {
                        matchingQueue.put(poison);
                    } catch (InterruptedException e) {
                        ExceptionHandler.handleExceptionExit(e);
                    }

                    done = true;
                }
            }

        }

        while (finishedThreads.get() < maxThread) {
            try {
                logger.info("Awaiting thread results");
                Thread.sleep(100);
            } catch (InterruptedException e) {
                logger.severe("InterruptedException");
                ExceptionHandler.handleExceptionExit(e);
            }
        }


        // Now we have to process the internal results into the format expected by the external results
        //

        logger.info("Reformatting results");
        for (int i = Global.MAX_SCORE; i > 0; i--) {
            List<Pair<Token, Token>> outerVec = internalResults.get(i);
            if (outerVec != null) {
                for (Pair<Token, Token> pair : outerVec) {
                    //System.err.println (i + ": " + pair.getA().name() + " | " + pair.getB().name());

                    List<String> updatedScores;
                    if (this.results.containsKey(i)) {
                        updatedScores = this.results.get(i);
                    } else {
                        updatedScores = new Vector<>();
                    }
                    updatedScores.add(String.format("%s|%s", pair.getA().name(), pair.getB().name()));
                    this.results.put(i, updatedScores);

                }
            }

        }
    }

}