package edu.georgetown.gucs.matcher;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.util.Comparator;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.Map.Entry;
import java.util.TreeMap;
import java.util.Vector;
import java.util.logging.Level;
import java.util.logging.Logger;


import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprint;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.fingerprinter.BulkFingerprinter;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.FileLister;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;
import org.apache.xerces.impl.dv.util.Base64;


/**
 * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are organized
 * by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if the
 * fingerprint does not already contain one) is optional.
 *
 * @author Lindsay Neubauer
 */
public class CompareFingerprint implements Iterable<Entry<Integer, List<String>>> {

    private Fingerprinter fingerprint;
    private final FingerprintMatcher matcher;
    private final FileLister fileLister;
    private final int minimum_score;
    private final Dictionary dictionary;
    private final SortedMap<Integer, List<String>> scores;
    private final int maxThread;
    private static final Logger logger = Logger.getLogger(BulkMatcher.class.getName());


    /**
     * Comparator that orders scores highest to lowest
     */
    private static class HightoLowScore implements Comparator<Integer>, Serializable {
        private static final long serialVersionUID = 5679507537454070748L;

        public HightoLowScore() {
        }

        @Override
        public int compare(Integer one, Integer two) {
            return two.compareTo(one);
        }
    }

    /**
     * Constructor that sets the matcher, fingerprint file, directory, minimum score and dictionary to use for this
     * comparison
     *
     * @param match           the string name of the matcher to use for this comparison
     * @param fingerprintFile the string filename of the fingerprint to compare with this directory
     * @param directory       the string directory path to compare with this fingerprint
     * @param minimumScore    the <code>int</code> minimum score needed to output a file for this comparison
     * @param dictionary      the string filename of the dictionary to use for this comparison; this dictionary will only be used if the
     *                        fingerprint file does not already contain a full dictionary
     *                        added global variables to set this.minimum_Score -- evan
     */
    public CompareFingerprint(String match, String fingerprintFile, String directory, int minimumScore, String dictionary) {

        this.matcher = (FingerprintMatcher) ObjectBuilder.buildObject(match);

        this.scores = Collections.synchronizedSortedMap(new TreeMap<>(new HightoLowScore()));

        this.fileLister = new FileLister(directory);
        this.fileLister.loadQueue();

        this.minimum_score = (minimumScore < Global.MIN_SCORE || minimumScore > Global.MAX_SCORE) ? Global.MIN_SCORE : minimumScore;

        //Determine if it is a N2 comparison
        if (fingerprintFile.contains("nSquared")) {
            this.dictionary = new Dictionary(dictionary);
        } else {
            this.fingerprint = new Fingerprinter(fingerprintFile, dictionary);
            this.dictionary = this.fingerprint.getDictionary();
        }

        this.maxThread = Runtime.getRuntime().availableProcessors();
    }

    @Override
    public Iterator<Entry<Integer, List<String>>> iterator() {
        return this.scores.entrySet().iterator();
    }

    /**
     * Compute a fingerprint for each file in this directory and compare it to this fingerprint using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     * <p>
     * added global variables to set {variable: score} --evan
     */
    //TODO This is broken because it calls the old computeFingerprintXML
    public void compareFingerprint() {

        logger.log(Level.INFO, "Directory", this.fileLister.getDirectory());
        Global.monitor.printMessage("CompareFingerprint", "Comparing fingerprint");

        Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(this.fingerprint.getFingerprintName());
        fingerprinter.setDictionary(this.dictionary);
        String matcherName = this.matcher.getClass().getSimpleName();
        int score;

        for (File file : this.fileLister) {
            try { //switched to computeFingerprint
                for (Fingerprint fingerprintPair : fingerprinter.computeFingerprint(file.getPath())) {

                    for (Fingerprint thisFingerprint : this.fingerprint.getBase64Fingerprints()) {

                        if (matcherName.equals("HashFingerprintMatcher") || matcherName.equals("ExactFingerprintMatcher")) {

                            score = (this.matcher.match(thisFingerprint.getBase64Fingerprint(),
                                    fingerprintPair.getBase64Fingerprint())) ? Global.MAX_SCORE : Global.MIN_SCORE;

                        } else {

                            score = this.matcher.getScoreXML(thisFingerprint.getBase64Fingerprint(),
                                    Base64.encode(fingerprintPair.getFingerprint()));
                        }

                        List<String> updatedScores;
                        if (this.scores.containsKey(score)) {
                            updatedScores = this.scores.get(score);
                        } else {
                            updatedScores = new Vector<String>();
                        }
                        updatedScores.add(fingerprintPair.getFileName());
                        this.scores.put(score, updatedScores);
                    }
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }


    /**
     * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     */
    public void compareFingerprintDirectory() {

        logger.info("Starting");

        List<Fingerprint> results;


        BulkFingerprinter bulkFP = new BulkFingerprinter(this.fingerprint.getFingerprintName(), this.fileLister, this.dictionary, maxThread);


        logger.log(Level.INFO, "Directory", this.fileLister.getDirectory());
        logger.info("Calling BulkFingerprinter");
        logger.info("Fingerprinting");

        System.err.println("Fingerprinting" + System.currentTimeMillis());

        results = bulkFP.computeAllFingerprints(this.fingerprint.getFingerprintName(), this.dictionary, this.fileLister);

        String matcherName = this.matcher.getClass().getSimpleName();

        logger.info("Calling BulkMatcher");

        logger.info("Matching");

        System.err.println("Matching " + System.currentTimeMillis());

        BulkMatcher bulkMatch = new BulkMatcher(matcherName, scores, maxThread);
        bulkMatch.compareAllFingerprints(results);

        logger.info("Matcher Complete");
    }

    /**
     * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     */
    public void compareFingerprintDirectory(String fingerprintVector) {
        logger.info("Starting");

        List<Fingerprint> results;


        BulkFingerprinter bulkFP = new BulkFingerprinter(fingerprintVector,
                this.fileLister, this.dictionary, maxThread);

        logger.log(Level.INFO, "Directory" + this.fileLister.getDirectory());
        logger.info("Calling BulkFingerprinter");
        logger.info("Fingerprinting");

        System.err.println("Fingerprinting" + System.currentTimeMillis());

        results = bulkFP.computeAllFingerprints(fingerprintVector, this.dictionary, this.fileLister);

        String matcherName = this.matcher.getClass().getSimpleName();

        logger.info("Calling BulkMatcher");
        logger.info("Matching");

        System.err.println("Matchings (CMPDir Fingerprint) "
                + System.currentTimeMillis());

        // Results contains all FP's
        // BulkMatcher now compares FP's
        // Edit so that if GAP is specified ...


        //if the matcher is googleAllPairsMatcher
        //store all the fingerprints
        Global.debug("matcher.getMatcherName()", matcher.getMatcherName());
        if (this.matcher.getMatcherName().contains("GoogleFingerprintMatcher")) {

            System.out.println("inside GAP Matcher ");
            //store all fingerprints
            matcher.match(results);

        } else {

            BulkMatcher bulkMatch = new BulkMatcher(matcherName, scores, maxThread);
            bulkMatch.compareAllFingerprints(results);
        }
        logger.info("Completed bulkMatcher");
    }


    /**
     * Compute a fingerprint for each file in this directory and compare it to each other using this matcher; if the
     * matcher returns boolean values, true values are given a score of 100 and false values are given a score of 0
     */
    public void googleAllcompareFpDir(String fingerprintVector) {
        logger.info("Starting GAP Compare");

        // initialize results
        List<Fingerprint> results;

        // initialize BLK fingerprinter with the fingerPrintVector to use,
        // fileList to iterate over,
        // dictionary of terms to match,
        // number of threads to create
        BulkFingerprinter bulkFP = new BulkFingerprinter(fingerprintVector,
                this.fileLister, this.dictionary, maxThread);

        logger.log(Level.INFO, "Directory" + this.fileLister.getDirectory());
        logger.info("Calling BulkFingerprinter");
        logger.info("Fingerprinting");

        // Run the fingerprinter on all files in the specified directory
        results = bulkFP.computeAllFingerprints(fingerprintVector, this.dictionary, this.fileLister);


        // Set the matcher used for comparisons
        String matcherName = this.matcher.getClass().getSimpleName();

        logger.info("Calling BulkMatcher");

        System.err.println("Matching: " + System.currentTimeMillis());

        BulkMatcher bulkMatch = new BulkMatcher(matcherName, scores, maxThread);

        // This is where we compare using googleAllPairs
        // Edit in bulkmatcherworker
        bulkMatch.compareAllFingerprints(results);

        logger.info("Completed matching all vectors");
    }

    /**
     * Provides all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     *
     * @return the string containing all the score results
     */
    private String getScores() {
        StringBuilder buffer = new StringBuilder();
        boolean printedSomething = false;
        for (int i = Global.MAX_SCORE; i >= this.minimum_score; i--) {
            if (this.scores.containsKey(i)) {
                for (String score : this.scores.get(i)) {
                    if (i < Global.MAX_SCORE && i > 10) buffer.append(String.format("%s%n", score + "|0" + i));
                    else if (i <= 10) buffer.append(String.format("%s%n", score + "|00" + i));
                    else buffer.append(String.format("%s%n", score + "|" + i));
                }
                printedSomething = true;
            }
        }
        if (!printedSomething) {
            buffer.append(String.format("%s",
                    "CompareDictionary: There are no files in the " +
                            "directory with a score higher than " + this.minimum_score));
        }
        return buffer.toString();
    }


    /**
     * Print all the files and their similarity scores; if a minimum score is provided, only output those files with
     * scores greater than or equal to the minimum
     */
    public void printScores() {
        System.out.println(getScores());
    }

    /**
     * Write all the files and their similarity scores to a file; if a minimum score is provided, only output those files
     * with scores greater than or equal to the minimum
     *
     * @param filename the string name of the file to write output
     */
    public void writeScores(String filename) {
        logger.log(Level.INFO, "Filename:", filename);
        try {
            OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(filename), Global.fileEncoding);
            BufferedWriter writer = new BufferedWriter(out);
            writer.write(getScores());
            writer.close();
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);
        }
        System.out.println("Writing scores to " + filename);
    }

    /**
     * Outputs a list of files and their scores, comparing a given fingerprint to a directory of files. Files are
     * organized by scores with higher scores appearing higher in the list. Providing a minimum score and a dictionary (if
     * the fingerprint does not already contain one) is optional.
     *
     * @param args array of string command line arguments<br>
     *             <code>args[0]</code> the name of the matcher to use for this comparison<br>
     *             <code>args[1]</code> the filename of the fingerprint to compare with this directory<br>
     *             <code>args[2]</code> the directory path to compare with this fingerprint<br>
     *             <code>args[3]</code> the optional integer minimum score needed to output a file for this comparison<br>
     *             <code>args[4]</code> the optional boolean indicating true if n-squared comparison desired, default: false<br>
     *             <code>args[5]</code> the optional filename of the dictionary to use for this comparison; only used if the
     *             fingerprint file does not contain a full dictionary or a link to a dictionary in the file system
     */
    public static void main(String args[]) {
        try {
            String matcher = args[0];
            String fingerprint = args[1];
            String directory = args[2];
            int min_score = (args.length > 3) ? Integer.valueOf(args[3]) : Global.MIN_SCORE;
            boolean n_squared = (args.length > 4) ? Boolean.valueOf(args[4]) : false;
            String dictionary = (args.length > 5) ? args[5] : null;

            CompareFingerprint compare = new CompareFingerprint(matcher, fingerprint, directory, min_score, dictionary);
            if (!n_squared) {
                compare.compareFingerprint();
            } else {
                compare.compareFingerprintDirectory(fingerprint);
            }
            compare.printScores();
            Global.monitor.closeBuffer();
        } catch (ArrayIndexOutOfBoundsException e) {
            System.err
                    .println("Usage: java CompareFingerprint <matcher name> <XML fingeprint file> <directory> <optional: minimum score> "
                            + "<optional: true/false for n-squared directory comparison> <optional: dictionaryFilename>");
            System.err
                    .println("Outputs a list of files and their scores, comparing a given fingerprint to a directory of files");
            System.exit(-1);
        } catch (RuntimeException exception) {
            ExceptionHandler.handleExceptionExit(exception);
        }
    }
}
