package edu.georgetown.gucs.tokenizers;

import java.io.*;
import java.util.*;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.Token;

/**
 * Uses Oracle's OutsideIn technology to extract text from a file. Splits contents of the file into tokens based on
 * whitespace or by line. Used for {@linkplain Dictionary} creation.
 *
 * @author Lindsay Neubauer
 */
public class OutsideInFileTokenizer extends FileTokenizer {

    private static final long serialVersionUID = -7648672098261208008L;

    /**
     * Constructor that sets the token creation mode to split based on whitespace
     */
    public OutsideInFileTokenizer() {
        this.mode = "tokens";
    }

    /**
     * Calls Oracle's Outside In API to extract text from a file and splits that text into tokens
     *
     */
    @Override
    protected List<Token> readFile(String filename) {
        File tempFile = null;
        try {
            InputStream inStream;
            BufferedReader buffer;

            String OS = System.getProperty("os.name").toLowerCase();

            tempFile = File.createTempFile("outsideInOutput", null);

            String libPath = Global.getProperty("libPath");
            String envVar = Global.getProperty("LD_LIBRARY_PATH");

            OutputStreamWriter writer = new OutputStreamWriter(new FileOutputStream(tempFile), Global.fileEncoding);
            if (OS.contains("win")) {

                inStream = new ProcessBuilder(libPath +"/outsideWin/MemoryIO.exe", filename)
                        .start().getInputStream();

                buffer = new BufferedReader(new InputStreamReader(inStream, Global.fileEncoding));

                String line;
                while ((line = buffer.readLine()) != null) {
                    writer.write(line);
                }

                inStream.close();

            } else {
                ProcessBuilder pb = new ProcessBuilder(libPath + "/outsideInExtract", filename);

                Map<String,String> env = pb.environment();

                env.put("LD_LIBRARY_PATH", envVar);

                Process p = pb.start();

                buffer = new BufferedReader(new InputStreamReader(p.getInputStream(), Global.fileEncoding));

                String line;
                while ((line = buffer.readLine()) != null) {
                    writer.write(line);
                }
            }
            writer.close();
            buffer.close();

            // FIXME OutsideInFileTokenizer: Does not separate out meta-tokens
            synchronized (this.tokenVectorMap) {
                this.tokenVectorMap.clear();

                Vector<Token> tokenVector = new Vector<>();

                synchronized (tokenVector) {
                    FileTokenizer fileTokenizer = new FileTokenizer(this.mode);

                    fileTokenizer.tokenize(tempFile.getAbsolutePath());

                    tokenVector.clear();

                    tokenVector.addAll(fileTokenizer.getTokenVectorMap().get("null"));

                    this.tokenVectorMap.put("null", tokenVector);

                    return tokenVector;
                }
            }
        } catch (IOException e) {
            ExceptionHandler.handleExceptionExit(e);

        } finally {
            if (tempFile != null) {
                tempFile.deleteOnExit();
            }
        }
        return null;
    }
}
