package sdtext;


import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Random;
import java.util.logging.*;
import java.util.zip.GZIPOutputStream;

import edu.georgetown.gucs.dictionary.Dictionary;
import edu.georgetown.gucs.dictionary.ShowDictionaryStatistics;
import edu.georgetown.gucs.dictionary.ShowDictionaryTokens;
import edu.georgetown.gucs.dictionary.TrimDictionary;
import edu.georgetown.gucs.experiment.Experiment;
import edu.georgetown.gucs.fingerprinter.BitVectorFingerprinter;
import edu.georgetown.gucs.fingerprinter.ExtractFingerprintDictionary;
import edu.georgetown.gucs.fingerprinter.SdhashFingerprinter;
import edu.georgetown.gucs.matcher.CompareFingerprint;
import edu.georgetown.gucs.matcher.ScoreFingerprints;
import edu.georgetown.gucs.matcher.SdHashFingerprintMatcher;
import edu.georgetown.gucs.tokenizers.ArabicFileTokenizer;
import edu.georgetown.gucs.tokenizers.ChineseFileTokenizer;
import edu.georgetown.gucs.tokenizers.TokenizeFile;
import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

/**
 * Provides a command line interface to run the different programs within SDText (Similarity Digest Text)
 *
 * @author Lindsay Neubauer
 */
public class CommandLine {

    private static String input = null; // -i, --input
    private static String output = null; // -o, --output
    private static String config = null; // -c, --config
    private static String dictionary = null; // -d, --dictionary
    private static String directory = null; // -p, --directory
    private static String fingerprint = null; // -f, --fingerprint
    private static String fingerprint2 = null;
    private static String matcher = null; // -m, --matcher
    private static double min_idf = 0.0; // -l, --min-idf
    private static double max_idf = 1.0; // -h, --max-idf
    private static int min_score = 0; // -s, --min-score
    private static final String basePath = Global.getProperty("basePath");

    /**
     * Prints SDText help screen, including all programs and their options
     */
    private static void printHelp() {
        boolean print_programs = false;
        for (CommandLineProgramsEnum program : CommandLineProgramsEnum.values()) {
            if (print_programs) {
                System.err.println();
                System.err.println("PROGRAMS:");
                System.err.println();
                print_programs = false;
            }
            System.err.println(program.name + ": " + program.use);
            System.err.println(program.options);
            System.err.println();
            if (program == CommandLineProgramsEnum.HELP)
                print_programs = true;
        }
    }

    /**
     * Prints program name, use, and command line options,
     * to be used when running a program that is missing a required
     * command line option
     *
     * @param program the program to print error message for
     */
    private static void printError(CommandLineProgramsEnum program) {
        String error = String.format("%s %n%s: %s %n%s%n", "SDText: Missing required OPTION", program.name, program.use,
                program.options);
        System.err.println(error);
        Global.monitor.writeToBuffer("CommandLine", "printError", error);
    }

    /**
     * Runs the specified SDText program
     *
     * @param program the program to execute
     */
    private static void sdtext(CommandLineProgramsEnum program) {
        switch (program) {
            case TOKENIZEFILE:
                if (input != null) {
                    if (config == null) {
                        config = basePath + "/doc/configuration_files/tokenizers_config.xml";
                        //   System.out.println(Global.getProperty("basePath"));
                    }
                    TokenizeFile tokenizeFile = new TokenizeFile(config);
                    tokenizeFile.tokenizeFile(input);
                    tokenizeFile.printTokens();
                } else {
                    printError(program);
                }
                break;
            case ARABICFILETOKENIZER:
                if (input != null) {
                    ArabicFileTokenizer arabicTokenizer = new ArabicFileTokenizer();
                    arabicTokenizer.tokenize(input);
                    arabicTokenizer.printTokens();
                } else {
                    printError(program);
                }
                break;
            case CHINESEFILETOKENIZER:
                if (input != null) {
                    ChineseFileTokenizer chineseTokenizer = new ChineseFileTokenizer();
                    chineseTokenizer.setMode("individual");
                    System.out.println("---------------------------");
                    System.out.println("Mode: Individual Characters");
                    System.out.println("---------------------------");
                    chineseTokenizer.tokenize(input);
                    chineseTokenizer.printTokens();
                    chineseTokenizer.setMode("smart");
                    System.out.println("------------------------------------------------");
                    System.out.println("Mode: Smart/Probabalistic Character Combinations");
                    System.out.println("------------------------------------------------");
                    chineseTokenizer.tokenize(input);
                    chineseTokenizer.printTokens();
                } else {
                    printError(program);
                }
                break;
            case CREATEDICTIONARY:
                if (directory != null) {
                    if (config == null) {
                        config = basePath + "/doc/configuration_files/tokenizers_config.xml";
                    }
                    Dictionary dict = new Dictionary();
                    dict.makeDictionary(config, directory, output);
                } else {
                    printError(program);
                }
                break;
            case SHOWDICTIONARYSTATISTICS:
                if (dictionary != null) {
                    ShowDictionaryStatistics.print(dictionary);
                } else {
                    printError(program);
                }
                break;
            case SHOWDICTIONARYTOKENS:
                if (dictionary != null) {
                    ShowDictionaryTokens.print(dictionary);
                } else {
                    printError(program);
                }
                break;
            case TRIMDICTIONARY:
                if (dictionary != null) {
                    if (output == null)
                        output = dictionary.substring(0, dictionary.indexOf(".")) + "_trimmed_" + min_idf + "-" + max_idf + ".xml";
                    TrimDictionary trimDictionary = new TrimDictionary(min_idf, max_idf, dictionary);
                    trimDictionary.trim(output);
                } else {
                    printError(program);
                }
                break;
            case BITVECTORFINGERPRINTER:
                System.out.println("dictionary " + dictionary + " " + input);
                if (dictionary != null && input != null) {
                    if (output == null) {
                        Random random = new Random();
                        output = "fingerprint_" + random.nextInt(Integer.MAX_VALUE) + ".xml";
                    }
                    BitVectorFingerprinter bvfp = new BitVectorFingerprinter(dictionary);
                    if (config != null) {
                        bvfp.outputFields(config);
                    }
                    bvfp.generateXML(bvfp.computeFingerprint(input), input, output);
                } else {
                    printError(program);
                }
                break;
            case SDHASHFINGERPRINTER:
                if (input != null) {

                    if (config == null) {
                        config = basePath + "/doc/configuration_files/tokenizers_config.xml";
                    }

                    SdhashFingerprinter sdFp = new SdhashFingerprinter();
                    sdFp.setTokenizers(config);
                    sdFp.computeFingerprint(input);
                    sdFp.printSdHashOutput();

                } else {
                    printError(program);
                }
                break;
            case SDHASHFINGERPRINTMATCHER:
                if (fingerprint != null && fingerprint2 != null) {

                    if (config == null) {
                        config = basePath + "/doc/configuration_files/tokenizers_config.xml";
                    }

                    SdhashFingerprinter sdFp = new SdhashFingerprinter();
                    SdHashFingerprintMatcher sdFpM = new SdHashFingerprintMatcher();

                    sdFp.setTokenizers(config);

                    sdFpM.getMatch(sdFp.computeFingerprint(fingerprint),
                            sdFp.computeFingerprint(fingerprint2));

                    sdFpM.printScore();
                } else {
                    printError(program);
                }
                break;

            case SCOREFINGERPRINTS:
                if (fingerprint != null && fingerprint2 != null) {

                    if (matcher == null) {
                        matcher = "CosineSimilarityFingerprintMatcher";
                    }

                    ScoreFingerprints sfp = new ScoreFingerprints(matcher, fingerprint, fingerprint2);
                    sfp.printMatch();

                } else {
                    printError(program);
                }
                break;
            case EXTRACTFINGERPRINTDICTIONARY:
                if (fingerprint != null) {
                    if (output == null)
                        ExtractFingerprintDictionary.extractDictionaryXML(fingerprint);
                    else
                        ExtractFingerprintDictionary.extractDictionaryXML(fingerprint, output);
                } else {
                    printError(program);
                }
                break;
            case COMPAREFINGERPRINT:    //Requires a fingerprint
                if (directory != null && fingerprint != null) {
                    if (matcher == null) matcher = "CosineSimilarityFingerprintMatcher";

                    CompareFingerprint compare = new CompareFingerprint(matcher, fingerprint, directory, min_score, dictionary);
                    compare.compareFingerprint();


                    if (output != null) {
                        compare.writeScores(output);
                    } else {
                        compare.printScores();
                    }
                } else {
                    printError(program);
                }
                break;
            case COMPAREDIRECTORY:  //fingerprint == bitvectorfingerprintMatcher
                if (directory != null) {
                    if (matcher == null) { matcher = "CosineSimilarityFingerprintMatcher"; }

                    CompareFingerprint compare;
                    if (fingerprint == null) {
                        fingerprint = "BitVectorFingerprinter";
                        String nSquared = "nSquared";
                        compare = new CompareFingerprint(matcher, fingerprint + nSquared, directory, min_score, dictionary);
                        compare.compareFingerprintDirectory(fingerprint);
                    } else {

                        compare = new CompareFingerprint(matcher, fingerprint, directory, min_score, dictionary);

                        compare.compareFingerprintDirectory();
                    }

                    if (output != null) {
                        compare.writeScores(output);
                    } else {
                        compare.printScores();
                    }
                } else {
                    printError(program);
                }
                break;
            case EXPERIMENT:

                // Check if a config file exists
                // If no config { error out }
                if (config != null) {
                    Experiment exp = new Experiment(config);

                    // Check if there is a name for the output file
                    // Otherwise print to stdOut
                    if (output != null) {
                        exp.setOutputFile(output);
                    }

                    // Check if temporary results file exists
                    File file;
                    if ((file = new File(Global.fileNameExperimentTracker)).isFile()) {
                        System.out.println(Global.fileNameExperimentTracker +
                                " deleted: " + file.delete());
                    }

                    // If fingerprint !null Store graph output filename in fingerprint
                    // Else use output fileName + .pdf
                    if (fingerprint == null) {
                        Random random = new Random();
                        fingerprint = "graph_" + random.nextInt(Integer.MAX_VALUE) ;
                    }

                    exp.runExperiment();

                    exp.printResults();

                    //set the name of the graph
                    exp.graphResults(fingerprint);

                } else {
                    printError(program);
                }
                break;
            default:
                break;
        }
    }

    /**
     * Provides a command line interface to run the different SDText programs
     *
     * @param args array of string command line arguments<br>
     *             <code>args[0]</code> the name of the program to run<br>
     *             <code>args[...]</code> the options needed to run the program
     */
    public static void main(String args[]) {
        // Check to see if we need to turn logging on
        Level level = Level.OFF;
        for (int i = 1; i < args.length; i += 2) {
            // if(args[i].trim().equalsIgnoreCase("-q") && (args[i+1].isEmpty() || args[i+1] == null))  printHelp();
            if (args[i].trim().equalsIgnoreCase("-q") || args[i].trim().equalsIgnoreCase("--debug")) {

                try {
                    level = Level.parse(args[i + 1].toUpperCase());
                } catch (ArrayIndexOutOfBoundsException e) {
                    System.err.println("ERROR: Logging Not turned on \n");
                    level = Level.OFF;
                } catch (IllegalArgumentException e) {
                    level = Level.OFF;
                    System.err.println("\n You must specify one of the following: ALL, Severe, info, warning \n");
                }
            }
        }
        Logger log = Logger.getLogger("");
        log.removeHandler(log.getHandlers()[0]);
        try {
            Handler handler = new StreamHandler(new GZIPOutputStream(new BufferedOutputStream(new FileOutputStream(
                    "log.gz"))), new SimpleFormatter()); //rotate?
            Logger.getLogger("").addHandler(handler);
        } catch (IOException e) {
            System.err.println("IOException: Unable to open stream for log results");
        }
        Handler[] handlers = Logger.getLogger("").getHandlers();
        for (int index = 0; index < handlers.length; index++) {
            handlers[index].setLevel(level);
        }


        try {
            if (Global.monitor.isLoggingOn()) {
                Runtime.getRuntime().addShutdownHook(new Thread() {
                    @Override
                    public void run() {
                        if (!Global.monitor.isClosed()) {
                            Global.monitor.writeToBuffer("CommandLine", "SDText", "Process killed");
                            Global.monitor.closeBuffer();
                        }
                    }
                });
            }

            String program = args[0].trim();
            if (!program.equalsIgnoreCase("-h") && !program.equalsIgnoreCase("--help")) {
                Global.monitor.writeToBuffer("CommandLine", "SDText", "program=" + args[0]);
            }


            // Now check the rest of the arguments and log them as needed
            for (int i = 1; i < args.length; i += 2) {

                if (args[i].trim().equalsIgnoreCase("-i") || args[i].trim().equalsIgnoreCase("--input")) {
                    input = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "input=" + input);

                } else if (args[i].trim().equalsIgnoreCase("-o") || args[i].trim().equalsIgnoreCase("--output")) {
                    output = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "output=" + output);

                } else if (args[i].trim().equalsIgnoreCase("-c") || args[i].trim().equalsIgnoreCase("--config")) {
                    config = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "config=" + config);

                } else if (args[i].trim().equalsIgnoreCase("-d") || args[i].trim().equalsIgnoreCase("--dictionary")) {
                    dictionary = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "dictionary=" + dictionary);

                } else if (args[i].trim().equalsIgnoreCase("-p") || args[i].trim().equalsIgnoreCase("--directory")) {
                    directory = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "directory=" + directory);

                } else if (args[i].trim().equalsIgnoreCase("-f") || args[i].trim().equalsIgnoreCase("--fingerprint")
                        || args[i].trim().equalsIgnoreCase("-g") || args[i].trim().equalsIgnoreCase("--sourceFingerprint")
                        || args[i].trim().equalsIgnoreCase("-t") || args[i].trim().equalsIgnoreCase("--targetFingerprint")
                        || args[i].trim().equalsIgnoreCase("-graph")) {


                    String errorArg = args[i].trim();

                    /* conditional checks that ensure parameters align with trial arguments
                      * the different trials */
                    if (program.equalsIgnoreCase("ScoreFingerprints")) {

                        if (errorArg.equals("-f") || errorArg.equals("--fingerprint")
                                || errorArg.equalsIgnoreCase("-v") || errorArg.equalsIgnoreCase("--fingerprintMatcher")) {

                            ExceptionHandler.handleExceptionExit(new Exception("SDText: Please use  " +
                                    "(-g OR --sourceFingerprint) and (-t OR --targetFingerprint)."));
                        }
                    }

                    if (program.equalsIgnoreCase("CompareFingerprint")) {

                        if (errorArg.equalsIgnoreCase("-g") || errorArg.equalsIgnoreCase("--sourcefingerprint")
                                || errorArg.equalsIgnoreCase("-t") || errorArg.equalsIgnoreCase("--targetfingerprint")) {
                            ExceptionHandler.handleExceptionExit(new Exception("SDText: Please use  " +
                                    "(-f OR --fingerprint with CompareFingerprint"));
                        }
                    }

                    if (program.equalsIgnoreCase("CompareDirectory")) {

                        if (errorArg.equalsIgnoreCase("-g") || errorArg.equalsIgnoreCase("--sourcefingerprint")
                                || errorArg.equalsIgnoreCase("-t") || errorArg.equalsIgnoreCase("--targetfingerprint")) {

                            ExceptionHandler.handleExceptionExit(new Exception("SDText: Please use  " +
                                    "(-v OR --fingerprintMatcher with CompareFingerprint"));
                        }

                        if (program.equalsIgnoreCase("experiment")) {

                            if (errorArg.equalsIgnoreCase("--sourcefingerprint") || errorArg.equalsIgnoreCase("-t")
                                    || errorArg.equalsIgnoreCase("--targetfingerprint")) {
                                ExceptionHandler.handleExceptionExit(new Exception("SDText: Please use  " +
                                        "(-f OR --fingerprint with CompareFingerprint"));
                            }
                        }


                    }

                    if (fingerprint == null) {
                        fingerprint = args[i + 1].trim();
                        Global.monitor.writeToBuffer("CommandLine", "SDText", "fingerprint=" + fingerprint);

                    } else if (fingerprint2 == null) {
                        fingerprint2 = args[i + 1].trim();
                        Global.monitor.writeToBuffer("CommandLine", "SDText", "fingerprint=" + fingerprint2);

                    } else {
                        ExceptionHandler.handleExceptionExit(new Exception("SDText: Too many fingerprints specified."));
                    }

                } else if (args[i].trim().equalsIgnoreCase("-m") || args[i].trim().equalsIgnoreCase("--matcher")) {
                    matcher = args[i + 1].trim();
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "matcher=" + matcher);

                } else if (args[i].trim().equalsIgnoreCase("-l") || args[i].trim().equalsIgnoreCase("--min-idf")) {
                    min_idf = Double.valueOf(args[i + 1]);
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "min_idf=" + min_idf);

                } else if (args[i].trim().equalsIgnoreCase("-h") || args[i].trim().equalsIgnoreCase("--max-idf")) {
                    max_idf = Double.valueOf(args[i + 1]);
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "max_idf=" + max_idf);

                } else if (args[i].trim().equalsIgnoreCase("-s") || args[i].trim().equalsIgnoreCase("--min-score")) {
                    min_score = Integer.valueOf(args[i + 1]);
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "min_score=" + min_score);

                } else if (args[i].trim().equalsIgnoreCase("-b") || args[i].trim().equalsIgnoreCase("--bool")) {
                    boolean bool = Boolean.valueOf(args[i + 1]);
                    Global.monitor.writeToBuffer("CommandLine", "SDText", "bool=" + bool);

                } else if (args[i].trim().equalsIgnoreCase("-q") || args[i].trim().equalsIgnoreCase("--debug")) {
                    // Skip this, as it was taken care of initially

                } else {
                    System.err.println("SDText=" + args[i] + " " + args[i + 1] + " is not a valid OPTION.");
                    System.err.println("Try 'sdtext --help' for more information.");
                    System.exit(-1);
                }
            }

            //@EVAN-- REDO THIS TO FIX NULL POINTER EXCEPTIONS
            if (program.equalsIgnoreCase("ArabicFileTokenizer")) {
                sdtext(CommandLineProgramsEnum.ARABICFILETOKENIZER);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("BitVectorFingerprinter")) {
                sdtext(CommandLineProgramsEnum.BITVECTORFINGERPRINTER);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("ChineseFileTokenizer")) {
                sdtext(CommandLineProgramsEnum.CHINESEFILETOKENIZER);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("SdHashfingerprinter")) {
                sdtext(CommandLineProgramsEnum.SDHASHFINGERPRINTER);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("CompareFingerprint")) {
                sdtext(CommandLineProgramsEnum.COMPAREFINGERPRINT);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("CompareDirectory")) {
                sdtext(CommandLineProgramsEnum.COMPAREDIRECTORY);  //EDIT HERE
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("CreateDictionary")) {
                sdtext(CommandLineProgramsEnum.CREATEDICTIONARY);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("Experiment")) {
                sdtext(CommandLineProgramsEnum.EXPERIMENT);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("ExtractFingerprintDictionary")) {
                sdtext(CommandLineProgramsEnum.EXTRACTFINGERPRINTDICTIONARY);
                Global.monitor.closeBuffer();

            } else if ("ScoreFingerprints".equalsIgnoreCase(program)) {
                sdtext(CommandLineProgramsEnum.SCOREFINGERPRINTS);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("ShowDictionaryStatistics")) {
                sdtext(CommandLineProgramsEnum.SHOWDICTIONARYSTATISTICS);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("ShowDictionaryTokens")) {
                sdtext(CommandLineProgramsEnum.SHOWDICTIONARYTOKENS);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("TokenizeFile")) {
                sdtext(CommandLineProgramsEnum.TOKENIZEFILE);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("TrimDictionary")) {
                sdtext(CommandLineProgramsEnum.TRIMDICTIONARY);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("SdHashFingerprintMatcher")) {
                sdtext(CommandLineProgramsEnum.SDHASHFINGERPRINTMATCHER);
                Global.monitor.closeBuffer();

            } else if (program.equalsIgnoreCase("-h") || program.equalsIgnoreCase("--help")) {
                printHelp();

            } else {
                System.err.println("SDText: " + program + " is not a valid PROGRAM.");
                printHelp();
            }
        } catch (ArrayIndexOutOfBoundsException e) {

            log.severe("ArrayOutOfBounds");
//            ExceptionHandler.handleExceptionExit(e);

            if (args.length == 0) {
                printHelp();
                System.exit(0);

            } else if (("CompareFingerprint").equalsIgnoreCase(args[0]))
                printError(CommandLineProgramsEnum.COMPAREFINGERPRINT);

            else if ("ArabicFileTokenizer".equalsIgnoreCase(args[0]))
                printError(CommandLineProgramsEnum.ARABICFILETOKENIZER);

            else if ("BitVectorFingerprinter".equalsIgnoreCase(args[0]))
                printError(CommandLineProgramsEnum.BITVECTORFINGERPRINTER);

            else if ("ChineseFileTokenizer".equalsIgnoreCase(args[0])) {
                printError(CommandLineProgramsEnum.CHINESEFILETOKENIZER);

            } else if (args[0].equalsIgnoreCase("CompareFingerprint")) {
                sdtext(CommandLineProgramsEnum.COMPAREFINGERPRINT);

            } else if (args[0].equalsIgnoreCase("CreateDictionary")) {
                printError(CommandLineProgramsEnum.CREATEDICTIONARY);

            } else if (args[0].equalsIgnoreCase("Experiment")) {
                printError(CommandLineProgramsEnum.EXPERIMENT);

            } else if (args[0].equalsIgnoreCase("ExtractFingerprintDictionary")) {
                printError(CommandLineProgramsEnum.EXTRACTFINGERPRINTDICTIONARY);

            } else if ("ScoreFingerprints".equalsIgnoreCase(args[0])) {
                printError(CommandLineProgramsEnum.SCOREFINGERPRINTS);
                System.out.println("[ScoreFP] error" + args[0]);

            } else if (args[0].equalsIgnoreCase("ShowDictionaryStatistics")) {
                printError(CommandLineProgramsEnum.SHOWDICTIONARYSTATISTICS);

            } else if (args[0].equalsIgnoreCase("ShowDictionaryTokens")) {
                printError(CommandLineProgramsEnum.SHOWDICTIONARYTOKENS);

            } else if (args[0].equalsIgnoreCase("TokenizeFile")) {
                printError(CommandLineProgramsEnum.TOKENIZEFILE);

            } else if (args[0].equalsIgnoreCase("TrimDictionary")) {
                printError(CommandLineProgramsEnum.TRIMDICTIONARY);

            } else if (args[0].equalsIgnoreCase("SdHashFingerprintMatcher")) {

                printError(CommandLineProgramsEnum.SDHASHFINGERPRINTMATCHER);
            }

        } catch (RuntimeException exception) {
            ExceptionHandler.handleExceptionExit(exception);
            exception.getMessage();
        }
    }
}
