package sdtext;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Arrays;

import javax.swing.GroupLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.LayoutStyle;
import javax.swing.ScrollPaneConstants;
import javax.swing.filechooser.FileNameExtensionFilter;

import edu.georgetown.gucs.fingerprinter.ExtractFingerprintDictionary;
import edu.georgetown.gucs.fingerprinter.Fingerprinter;
import edu.georgetown.gucs.utility.Global;
import edu.georgetown.gucs.utility.ObjectBuilder;

/**
 * Provides an interface for receiving action events related to a fingerprinter; used with {@linkplain GUI}.
 * 
 * @author Lindsay Neubauer
 * 
 */
class FingerprinterActionListener implements ActionListener {

  private final JCheckBox digestBox = new JCheckBox("Include Digest");
  private final JCheckBox dictionaryBox = new JCheckBox("Include Dictionary");
  private final JCheckBox dataSourceBox = new JCheckBox("Include Data Source");
  private final JComboBox fingerprinterChooser = new JComboBox();
  private JButton dictionaryLink = new JButton("");
  private final JButton fileLink = new JButton("");
  private final JButton directoryLink = new JButton("");
  private final JButton fingerprintLink = new JButton("");
  final JButton fingerprintButton = new JButton("");
  private final JTextArea fingerprintOutput = new JTextArea("");
  private final boolean create;

  FingerprinterActionListener(boolean create) {
    this.create = create;
    GUI.linkSetup(this.dictionaryLink);
    GUI.linkSetup(this.fileLink);
    GUI.linkSetup(this.directoryLink);
    GUI.linkSetup(this.fingerprintLink);

    if (this.create) {
      this.fingerprintButton.setText("Create Fingerprint");
    } else {
      this.fingerprintButton.setText("Extract Dictionary");
    }
    this.fingerprintButton.addActionListener(this);
    this.fingerprintButton.setEnabled(false);

    for (FingerprinterEnum fingerprinter : FingerprinterEnum.values()) {
      this.fingerprinterChooser.addItem(fingerprinter);
    }
    this.fingerprinterChooser.setSelectedIndex(0);
    this.fingerprinterChooser.setMaximumSize(new Dimension(200, 50));

    this.digestBox.setSelected(true);
    this.dictionaryBox.setSelected(true);
    this.dataSourceBox.setSelected(false);

    this.fingerprintOutput.setMargin(new Insets(5, 5, 5, 5));
    this.fingerprintOutput.setEditable(false);
  }

  /**
   * Provides the link to the dictionary file to use for fingerprinting
   * 
   * @return the JButton containing the link to the dictionary file
   */
  public JButton getDictionaryLink() {
    return this.dictionaryLink;
  }

  /**
   * Sets the link to the dictionary file to use for fingerprinting
   * 
   * @param dictionaryLink
   *          the JButton containing the link to the dictionary file
   */
  public void setDictionaryLink(JButton dictionaryLink) {
    this.dictionaryLink = dictionaryLink;
  }

  @Override
  public void actionPerformed(ActionEvent actionEvent) {
    ((JButton) actionEvent.getSource()).setEnabled(false);
    if (this.create) {
      this.fingerprintOutput.setText("");
      if (this.dictionaryLink.getText().length() > 0
          && GUI.xmlFileTypeCheck("dictionary", this.dictionaryLink.getActionCommand())) {
        if (this.fileLink.getText().length() > 0) {
          this.fingerprintLink.setText("Creating Fingerprint(s)...");
          Fingerprinter fingerprinter = (Fingerprinter) ObjectBuilder
              .buildObject(((FingerprinterEnum) this.fingerprinterChooser.getSelectedItem()).name);
          fingerprinter.setOutput(this.digestBox.isSelected(), this.dictionaryBox.isSelected(),
              this.dataSourceBox.isSelected());
          fingerprinter.setDictionary(this.dictionaryLink.getActionCommand());
          JFileChooser fileChooser = new JFileChooser();
          fileChooser.setFileFilter(new FileNameExtensionFilter("XML Files", "xml"));
          fileChooser.setDialogTitle("Save as XML Fingerprint File");
          int returnVal = fileChooser.showSaveDialog((Component) actionEvent.getSource());
          if (returnVal == JFileChooser.APPROVE_OPTION) {
            try {
              String filename = (fileChooser.getSelectedFile().getAbsolutePath().contains(".xml")) ? fileChooser
                  .getSelectedFile().getAbsolutePath() : fileChooser.getSelectedFile().getAbsolutePath() + ".xml";

              this.fingerprintOutput.append(fingerprinter.generateXML(
                  fingerprinter.computeFingerprintXML(this.fileLink.getActionCommand()),
                  this.directoryLink.getActionCommand(), filename));
              GUI.createLink(filename, this.fingerprintLink, true);
            } catch (RuntimeException exception) {
              this.fingerprintLink.setText("");
              Global.monitor.printErrorMessage("FingerprinterActionListener", "Exception: " + exception.getMessage());
              Global.monitor.writeToBuffer("FingerprinterActionListener", "fingerprintButton.actionPerformed",
                  exception.getMessage());
              Global.monitor.writeToBuffer("FingerprinterActionListener", "fingerprintButton.actionPerformed",
                  Arrays.toString(exception.getStackTrace()));
              JOptionPane.showConfirmDialog((Component) actionEvent.getSource(),
                  "Exception: " + exception.getMessage(), "Error: Fingerprinter", JOptionPane.CANCEL_OPTION,
                  JOptionPane.ERROR_MESSAGE);
            }
          } else {
            this.fingerprintLink.setText("");
          }
        } else {
          JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), "Must select a file to fingerprint",
              "Error: Fingerprinter", JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
        }
      } else {
        JOptionPane.showConfirmDialog((Component) actionEvent.getSource(),
            "Must load the existing dictionary before fingerprinting", "Error: Fingerprinter",
            JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
      }
    } else {
      if (this.fingerprintLink.getText().length() > 0) {
        this.dictionaryLink.setText("<HTML>Extracting Dictionary...</HTML>");
        JFileChooser fileChooser = new JFileChooser();
        fileChooser.setFileFilter(new FileNameExtensionFilter("XML Files", "xml"));
        fileChooser.setDialogTitle("Save as XML Dictionary File");
        int returnVal = fileChooser.showSaveDialog((Component) actionEvent.getSource());
        if (returnVal == JFileChooser.APPROVE_OPTION) {
          ExtractFingerprintDictionary.extractDictionaryXML(this.fingerprintLink.getActionCommand(), fileChooser
              .getSelectedFile().getAbsolutePath());
          GUI.createLink(fileChooser.getSelectedFile().getAbsolutePath(), this.dictionaryLink, false);
        } else {
          this.dictionaryLink.setText("");
        }
      } else {
        JOptionPane.showConfirmDialog((Component) actionEvent.getSource(),
            "Must load the existing dictionary before fingerprinting", "Error: Fingerprinter",
            JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
      }
    }
    ((JButton) actionEvent.getSource()).setEnabled(true);
  }

  /**
   * Creates a panel containing all the necessary components for extracting a dictionary from a fingerprint
   * 
   * @param layout
   *          the GroupLayout to add necessary dictionary extraction components
   */
  protected void extractDictionaryPanel(GroupLayout layout) {
    JPanel dictionaryPanel = GUI.createSubPanel(
        Arrays.asList(new JComponent[] { this.dictionaryLink, this.fingerprintButton }), "Extract Dictionary", true);
    dictionaryPanel.setMinimumSize(new Dimension(600, 100));
    dictionaryPanel.setMaximumSize(new Dimension(5000, 100));
    JPanel fingerprintPanel = createLoadPanel("Fingerprint File", this.fingerprintLink, this.fingerprintButton,
        Global.XML_ONLY);

    layout
        .setHorizontalGroup(layout.createParallelGroup().addComponent(fingerprintPanel).addComponent(dictionaryPanel));
    layout
        .setVerticalGroup(layout.createSequentialGroup().addComponent(fingerprintPanel).addComponent(dictionaryPanel));
  }

  /**
   * Creates a panel containing all the necessary components for creating a fingerprint
   * 
   * @param layout
   *          the GroupLayout to add necessary fingerprint creation components
   */
  protected void createFingerprintPanel(GroupLayout layout) {
    JPanel fingerprinterChooserPanel = GUI.createSubPanel(
        Arrays.asList(new JComponent[] { this.fingerprinterChooser }), "Fingerprinter", true);
    JPanel dictionaryPanel = createLoadPanel("Existing Dictionary", this.dictionaryLink, null, Global.XML_ONLY);
    JPanel filePanel = createLoadPanel("File to Fingerprint", this.fileLink, this.fingerprintButton, Global.ALL_FILES);
    JScrollPane scrollPane = new JScrollPane(this.fingerprintOutput, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    scrollPane.setMinimumSize(new Dimension(75, 200));

    // FIXME FingerprintActionListener: Add digestBox and dataSourceBox once implemented in Fingerprinter
    JPanel fingerprintPanel = GUI.createSubPanel(
        Arrays.asList(new JComponent[] {
            GUI.createSubPanel(Arrays.asList(new JComponent[] { this.dictionaryBox }), "Output Options", false),
            this.fingerprintLink, this.fingerprintButton, scrollPane }), "Create Fingerprint", true);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(fingerprinterChooserPanel)
        .addComponent(dictionaryPanel).addComponent(filePanel).addComponent(fingerprintPanel));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(fingerprinterChooserPanel)
        .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED).addComponent(dictionaryPanel)
        .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED).addComponent(filePanel)
        .addPreferredGap(LayoutStyle.ComponentPlacement.RELATED).addComponent(fingerprintPanel));
  }

  /**
   * Provides a panel that allows a user to load a file or directory
   * 
   * @param panelLabel
   *          string name of panel to create
   * @param panelLink
   *          JButton link to store the loaded file or directory
   * @param buttonToEnable
   *          JButton to enable once panel link is loaded; can be null
   * @param showAll
   *          boolean value indicating which files to display (Global.ALL_FILES or Global.XML_ONLY)
   * 
   * @return the JPanel containing the label, link, and bottom that allows user to load file/directory
   */
  private JPanel createLoadPanel(final String panelLabel, final JButton panelLink, final JButton buttonToEnable,
      final boolean showAll) {

    JLabel label = new JLabel(panelLabel);
    label.setFont(Global.boldUnifontFont);
    final JButton directoryButton = new JButton("Select " + panelLabel);
    directoryButton.setFont((new JLabel()).getFont());
    directoryButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent actionEvent) {
        directoryButton.setEnabled(false);
        panelLink.setText("<HTML>Loading " + panelLabel + "...</HTML>");
        JFileChooser chooser = new JFileChooser();
        if (!showAll) {
          chooser.setFileFilter(new FileNameExtensionFilter("XML Files", "xml"));
        }
        chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        chooser.setDialogTitle("Choose a " + panelLabel);
        int returnVal = chooser.showOpenDialog((Component) actionEvent.getSource());
        if (returnVal == JFileChooser.APPROVE_OPTION) {
          if (panelLabel.contains("Dictionary")
              && !GUI.xmlFileTypeCheck("dictionary", chooser.getSelectedFile().getAbsolutePath())) {
            panelLink.setText("");
            panelLink.setActionCommand("");
            JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), "Must load existing dictionary",
                "Error: Dictionary", JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
          } else if (panelLabel.contains("Fingerprint File")
              && !GUI.xmlFileTypeCheck("fingerprint", chooser.getSelectedFile().getAbsolutePath())) {
            panelLink.setText("");
            panelLink.setActionCommand("");
            JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), "Must load existing fingerprint",
                "Error: Fingerprinter", JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
          } else {
            panelLink.setActionCommand(chooser.getSelectedFile().getAbsolutePath());
            GUI.createLink(chooser.getSelectedFile().getAbsolutePath(), panelLink, false);
            if (FingerprinterActionListener.this.fingerprintButton != null) {
              FingerprinterActionListener.this.fingerprintButton.setEnabled(true);
            }
          }
        } else {
          panelLink.setText("");
          panelLink.setActionCommand("");
        }
        directoryButton.setEnabled(true);
      }
    });

    JPanel panel = new JPanel();
    GroupLayout layout = GUI.createGroupLayout(panel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(label).addComponent(panelLink)
        .addComponent(directoryButton));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(label).addComponent(panelLink)
        .addComponent(directoryButton).addPreferredGap(label, panelLink, LayoutStyle.ComponentPlacement.INDENT)
        .addPreferredGap(label, directoryButton, LayoutStyle.ComponentPlacement.INDENT));
    return panel;
  }
}
