package sdtext;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.GroupLayout;
import javax.swing.GroupLayout.ParallelGroup;
import javax.swing.GroupLayout.SequentialGroup;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.LayoutStyle;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

/**
 * Provides a graphical user interface to run the different programs within SDText (Similarity Digest Text)
 * 
 * @author Lindsay Neubauer
 * 
 */
class GUI extends JPanel implements ChangeListener {

  private static final long serialVersionUID = -2132114502654884184L;
  private static DictionaryActionListener existingDictionaryListener;
  private static FingerprinterActionListener createFingerprinterListener;

  /**
   * Constructor that sets up the GUI to run the different SDText programs
   */
  public GUI() {
    super(new GridLayout(1, 1));

    JTabbedPane tabbedPane = new JTabbedPane();
    tabbedPane.addChangeListener(this);
    tabbedPane.addTab("Instructions", makeInstructionPanel());
    tabbedPane.addTab("Tokenize", makeTokenizePanel());
    tabbedPane.addTab("Dictionary", makeDictionaryPanel());
    tabbedPane.addTab("Fingerprint", makeFingerprintPanel());
    tabbedPane.addTab("Matcher", makeMatcherPanel());
    add(tabbedPane);
    tabbedPane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
  }

  @Override
  public void stateChanged(ChangeEvent changeEvent) {
    Global.monitor.writeToBuffer("GUI", "TabChangeListener", ((JTabbedPane) changeEvent.getSource())
        .getSelectedComponent().getName());
  }

  private static JComponent makeInstructionPanel() {
    JPanel tokenizePanel = new JPanel();
    JLabel tokenizeLabel = new JLabel("Tokenize");
    tokenizeLabel.setFont(Global.boldUnifontFont14);
    JLabel tokenizeDirections = new JLabel();
    tokenizeDirections.setText("<HTML>Splits a file into tokens and displays the resulting tokens<OL>"
        + "<LI>Choose a file tokenizer from the drop down menu. Select any additional options that <BR>"
        + "may appear as radio buttons below the drop down menu after a new file tokenizer is <BR>" + "selected.</LI>"
        + "<LI>Drag additional tokenizers from the <I>Tokenizer Options</I> list to the <I>Selected Tokenizers</I><BR>"
        + "list to further tokenize the file. Tokenizers in the <I>Selected Tokenizers</I> list may be<BR>"
        + "reordered. The order of these tokenizers matters.</LI>"
        + "<LI>Click <I>Create Dictionary</I> button to select a text file from the local file system to tokenize.<BR>"
        + "Tokens will appear in the textbox below the button.</LI></OL></HTML>");
    GroupLayout layout = createGroupLayout(tokenizePanel);
    layout
        .setHorizontalGroup(layout.createParallelGroup().addComponent(tokenizeLabel).addComponent(tokenizeDirections));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(tokenizeLabel)
        .addPreferredGap(tokenizeLabel, tokenizeDirections, LayoutStyle.ComponentPlacement.UNRELATED)
        .addComponent(tokenizeDirections));

    JPanel dictionaryPanel = new JPanel();
    JLabel dictionaryLabel = new JLabel("Dictionary");
    dictionaryLabel.setFont(Global.boldUnifontFont14);
    JLabel dictionaryDirections = new JLabel();
    dictionaryDirections.setText("<HTML>Creates a dictionary<OL>"
        + "<LI>Choose a file tokenizer from the drop down menu. Select any additional options that <BR>"
        + "may appear as radio buttons below the drop down menu after a new file tokenizer is <BR>" + "selected.</LI>"
        + "<LI>Drag additional tokenizers from the <I>Tokenizer Options</I> list to the <I>Selected Tokenizers</I><BR>"
        + "list to further tokenize the file. Tokenizers in the <I>Selected Tokenizers</I> list may be<BR>"
        + "reordered. The order of these tokenizers matters.</LI>"
        + "<LI>Move each end of the double-sided slider to specify minimum and maximum normalized <BR>"
        + "inverse document frequencies (IDFs) to include in the dictionary. By default these are set <BR>"
        + "to 0 and 1, respectively.</LI>"
        + "<LI>Click <I>Select Directory</I> button to select a directory containing text files to use to create <BR>"
        + "a dictionary.</LI>"
        + "<LI>Click <I>Create Dictionary</I> button to choose the name and location to create and save <BR>"
        + "dictionary.</OL>"

        + "Loads an existing dictionary<OL>"
        + "<LI>Click the <I>Select Existing Dictionary</I> button to select an existing dictionary XML file from <BR>"
        + "the local file system.</LI>"
        + "<LI>Click the <I>Show Dictionary Statistics</I> button to display the dictionary's:<UL>"
        + "<LI>Name</LI><LI>Language</LI><LI>Number of documents</LI><LI>Number of tokens</LI>"
        + "<LI>If the dictionary has been trimmed (including the IDF range, if trimmed)</LI>"
        + "<LI>Minimum and maximum IDF</LI><LI>List of tokenizers</LI></UL></LI>"
        + "<LI>Click the <I>Show Dictionary Tokens</I> button to display all the tokens and their frequencies, <BR>"
        + "IDFs, and normalized IDFs contained in the dictionary.</LI>"
        + "<LI>Trim the existing dictionary by moving either end of the double-sided slider to specify <BR>"
        + "the new minimum and maximum normalized inverse document frequencies (IDFs)  to <BR>"
        + "include in the trimmed dictionary. By default these are set to 0.3 and 0.7, respectively. <BR>"
        + "Click the <I>Trim Dictionary</I> button to choose the name and location to create and <BR>"
        + "save dictionary.</LI></OL></HTML>");
    layout = createGroupLayout(dictionaryPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(dictionaryLabel)
        .addComponent(dictionaryDirections));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(dictionaryLabel)
        .addPreferredGap(dictionaryLabel, dictionaryDirections, LayoutStyle.ComponentPlacement.UNRELATED)
        .addComponent(dictionaryDirections));

    JPanel fingerprintPanel = new JPanel();
    JLabel fingerprintLabel = new JLabel("Fingerprint");
    fingerprintLabel.setFont(Global.boldUnifontFont14);
    layout = createGroupLayout(fingerprintPanel);
    JLabel fingerprintDirections = new JLabel();
    fingerprintDirections
        .setText("<HTML>Creates a fingerprint<OL>"
            + "<LI>Choose a fingerprinter from the drop down menu.</LI>"
            + "<LI>Click the <I>Select Existing Dictionary</I> button to select an existing dictionary XML file from <BR>"
            + "the local file system.</LI>"
            + "<LI>Click the <I>Select File to Fingerprint</I> button to select a text file from the local file system <BR>"
            + "to fingerprint. The file must be compatible with the Dictionary's FileTokenizer.</LI>"
            + "<LI>Check or un-check whether to include the full dictionary in the fingerprint output file. <BR>"
            + "Check box is selected by default.</LI>"
            + "<LI>Click the <I>Create Fingerprint</I> button to choose the name and location to create and save <BR>"
            + "the fingerprint output file. Once the fingerprint output file is saved, the computed <BR>"
            + "Base64 encoded fingerprint will appear in the textbox below the button.</LI></OL>"

            + "Extracts a dictionary contained in a fingerprint<OL>"
            + "<LI>Click the <I>Select Fingerprint File</I> button to select an existing fingerprint XML file from <BR>"
            + "the local file system.</LI>"
            + "<LI>Click the <I>Extract Dictionary</I> button to choose the name and location to save the <BR>"
            + "extracted dictionary to the file system.</LI></OL></HTML>");
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(fingerprintLabel)
        .addComponent(fingerprintDirections));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(fingerprintLabel)
        .addPreferredGap(fingerprintLabel, fingerprintDirections, LayoutStyle.ComponentPlacement.UNRELATED)
        .addComponent(fingerprintDirections));

    JPanel matcherPanel = new JPanel();
    JLabel matcherLabel = new JLabel("Matcher");
    matcherLabel.setFont(Global.boldUnifontFont14);
    JLabel matcherDirections = new JLabel();
    matcherDirections.setText("<HTML>Scores two fingerprints<OL>"
        + "<LI>Choose a fingerprint matcher from the drop down menu.</LI>"
        + "<LI>Click the <I>Select First Fingerprint</I> button to select an existing fingerprint XML file from <BR>"
        + "the local file system.</LI>"
        + "<LI>Click the <I>Select Second Fingerprint</I> button to select an existing fingerprint XML file <BR>"
        + "from the local file system.</LI>"
        + "<LI>Click the <I>Compare Fingerprints</I> button to display the match results for the two <BR>"
        + "fingerprints.</OL>"

        + "Compares a fingerprint to all the files in a directory<OL>"
        + "<LI>Choose a fingerprint matcher from the drop down menu. The <BR>"
        + "<I>CosineSimilarityFingerprintMatcher</I> has a minimum score option that will only display <BR>"
        + "fingerprint matches above the given score. By default this score is set to 0.</LI>"
        + "<LI>Click the <I>Select Fingerprint</I> button to select an existing fingerprint XML file from the <BR>"
        + "local file system.</LI>"
        + "<LI>Click the <I>Select Directory</I> button to select a directory of text files to compare to the <BR>"
        + "existing fingerprint."
        + "<LI>Click the <I>Select Dictionary</I> button to select an existing dictionary XML file to use for the <BR>"
        + "comparison. If the existing " + "fingerprint contains a dictionary or a reference to <BR>"
        + "a dictionary on the local file system, an additional dictionary does not need to be <BR>" + "selected.</LI>"
        + "<LI>Click the <I>Compare to Directory</I> button to display the match results for each file in the <BR>"
        + "directory to the fingerprint. " + "Files are organized by scores with higher scores appearing <BR>"
        + "higher in the list</OL></HTML>");
    layout = createGroupLayout(matcherPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(matcherLabel).addComponent(matcherDirections));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(matcherLabel)
        .addPreferredGap(matcherLabel, matcherDirections, LayoutStyle.ComponentPlacement.UNRELATED)
        .addComponent(matcherDirections));

    JPanel experimentPanel = new JPanel();
    JLabel experimentLabel = new JLabel("Experiment");
    experimentLabel.setFont(Global.boldUnifontFont14);
    JLabel experimentDirections = new JLabel();
    experimentDirections.setText("<HTML>Create Configuration File and Run Experiment<OL>"
        + "<LI>Select groups of tokenizers and manglers<UL>"
        + "<LI>Choose a file tokenizer from the drop down menu. Select any additional options <BR>"
        + "that may appear as radio buttons below the drop down menu after a new file <BR>"
        + "tokenizer is selected.</LI>"
        + "<LI>Drag any manglers from the <I>Mangler Options</I> list to the <I>Selected Manglers</I> list <BR>"
        + "to add different file manipulations to test robustness to edits. Manglers in the <BR>"
        + "<I>Selected Manglers</I> list may be reordered. The order of these manglers matters. <BR>"
        + "Manglers are not required to run an experiment.</LI>"
        + "<LI>Drag additional tokenizers from the <I>Tokenizer Options</I> list to the <BR>"
        + "<I>Selected Tokenizers</I> list to further tokenize the files. Tokenizers in the <BR>"
        + "<I>Selected Tokenizers</I> list may be reordered. The order of these tokenizers<BR>"
        + "matters. These additional tokenizers are not required to run an experiment.</LI>"
        + "<LI>Click the <I>Add Tokenizers</I> button to add a group of tokenizers and manglers to <BR>"
        + "the experiment. Multiple groups may be added to run an experiment but <BR>"
        + "at least one group must be chosen. To remove a group of tokenizers and <BR>"
        + "manglers, select it and click the <I>Delete Tokenizers</I> button.</LI></UL></LI>"

        + "<LI>Drag at least one fingerprinter from the <I>Fingerprinter Options</I> list to the<BR>"
        + "<I>Selected Fingerprinter</I> list.</LI>"
        + "<LI>Drag at least one matcher from the <I>Matcher Options</I> list to the <I>Selected Matcher</I> list.<BR>"
        + "The <I>CosineSimilarityFingerprintMatcher</I> has a minimum score option that will only <BR>"
        + "consider fingerprint matches above the given score. By default this score is set to 0.</LI>"
        + "<LI>Click <I>Select Dataset Directory</I> button to select a directory containing text files to use <BR>"
        + "for dictionary creation.</LI>"
        + "<LI>Choose the language of the dataset files from the drop down menu.</LI>"
        + "<LI>Set the dictionary inverse document frequency (IDF) range by moving either end of the <BR>"
        + "double-sided slider to specify the minimum and maximum normalized IDFs. Click the <BR>"
        + "<I>Add IDF Range</I> button to add the selected minimum and maximum IDF to the <BR>"
        + "experiment. Multiple IDF ranges may be added to run an experiment. To remove an IDF <BR>"
        + "range, select it and click the <I>Delete IDF Range</I> button.</LI>"
        + "<LI>Choose <I>Percent Files</I> or <I>File Count</I> radio button and specify the percentage or number <BR>"
        + "of files to use for the dictionary size parameter. This should be low (&#60; 25%) so there are <BR>"
        + "enough files for testing against (sample files).</LI>"
        + "<LI>Choose <I>Percent Files</I> or <I>File Count</I> radio button and specify the percentage or number <BR>"
        + "of files to use for the sample size parameter. The files are chosen randomly from the <BR>"
        + "collection of files not already used to create the dictionary.</LI>"
        + "<LI>Click <I>Select Plots Directory</I> button to select the directory to save PDF result plots.</LI>"
        + "<LI>Check which result plots to create. By default <I>Precision vs. Recall</I> is selected.</LI>"
        + "<LI>Click <I>Run Experiment</I> button to choose the name and location to create and save the <BR>"
        + "XML experiment configuration file. This button will also run an experiment with <BR>"
        + "the given parameters.</LI>" + "</OL></HTML>");
    layout = createGroupLayout(experimentPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(experimentLabel)
        .addComponent(experimentDirections));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(experimentLabel)
        .addPreferredGap(experimentLabel, experimentDirections, LayoutStyle.ComponentPlacement.UNRELATED)
        .addComponent(experimentDirections));

    JPanel instructionPanel = new JPanel();
    instructionPanel.setName("InstructionPanel");
    JScrollPane instructionScroll = new JScrollPane(instructionPanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
    instructionScroll.setMinimumSize(new Dimension(100, 100));
    layout = createGroupLayout(instructionPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(tokenizePanel).addComponent(dictionaryPanel)
        .addComponent(fingerprintPanel).addComponent(matcherPanel).addComponent(experimentPanel));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(tokenizePanel).addComponent(dictionaryPanel)
        .addComponent(fingerprintPanel).addComponent(matcherPanel).addComponent(experimentPanel));
    return instructionScroll;
  }

  private static JComponent makeTokenizePanel() {
    // Select tokenizers
    JPanel selectionPanel = new JPanel();
    TokenizerSelectionHandler tokenizerHandler = new TokenizerSelectionHandler();
    tokenizerHandler.createPanel(createGroupLayout(selectionPanel), false, false);

    // Select file and tokenize it
    TokenizeFileActionListener tokenizeFileActionListener = new TokenizeFileActionListener(tokenizerHandler);
    JPanel tokenizeFile = tokenizeFileActionListener.createPanel();

    JPanel tokenizePanel = new JPanel();
    tokenizePanel.setName("TokenizePanel");
    tokenizePanel.setBorder(BorderFactory.createTitledBorder("Tokenize a File"));

    GroupLayout layout = createGroupLayout(tokenizePanel);
    layout.setHorizontalGroup(layout.createParallelGroup(GroupLayout.Alignment.CENTER).addComponent(selectionPanel)
        .addComponent(tokenizeFile));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(selectionPanel).addComponent(tokenizeFile));
    return new JScrollPane(tokenizePanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
  }

  private static JComponent makeDictionaryPanel() {
    // Create new dictionary
    JPanel newDictPanel = new JPanel();
    newDictPanel.setBorder(BorderFactory.createTitledBorder("Create New Dictionary"));
    DictionaryActionListener dictionaryListener = new DictionaryActionListener("create");
    dictionaryListener.createNewDictionaryPanel(createGroupLayout(newDictPanel));

    // Load existing dictionary
    JPanel existingDictPanel = new JPanel();
    existingDictPanel.setBorder(BorderFactory.createTitledBorder("Load Existing Dictionary"));
    existingDictionaryListener = new DictionaryActionListener("trim");
    existingDictionaryListener.createLoadDictionaryPanel(createGroupLayout(existingDictPanel));

    JPanel dictionaryPanel = new JPanel();
    dictionaryPanel.setName("DictionaryPanel");
    GroupLayout layout = createGroupLayout(dictionaryPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(newDictPanel).addComponent(existingDictPanel));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(newDictPanel).addComponent(existingDictPanel));
    return new JScrollPane(dictionaryPanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
  }

  private static JComponent makeFingerprintPanel() {
    // Create Fingerprint
    JPanel createPanel = new JPanel();
    createPanel.setBorder(BorderFactory.createTitledBorder("Create Fingerprint"));
    createFingerprinterListener = new FingerprinterActionListener(true);
    createFingerprinterListener.createFingerprintPanel(createGroupLayout(createPanel));

    // Extract Dictionary from Fingerprint File
    JPanel loadPanel = new JPanel();
    loadPanel.setBorder(BorderFactory.createTitledBorder("Extract Dictionary from Fingerprint"));
    FingerprinterActionListener extractFingerprinterListener = new FingerprinterActionListener(false);
    extractFingerprinterListener.extractDictionaryPanel(createGroupLayout(loadPanel));

    JPanel fingerprinterPanel = new JPanel();
    fingerprinterPanel.setName("FingerprinterPanel");
    GroupLayout layout = createGroupLayout(fingerprinterPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(createPanel).addComponent(loadPanel));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(createPanel).addComponent(loadPanel));
    return new JScrollPane(fingerprinterPanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
  }

  private static JComponent makeMatcherPanel() {
    // Score Fingerprints
    JPanel scorePanel = new JPanel();
    scorePanel.setBorder(BorderFactory.createTitledBorder("Score Fingerprints"));
    MatcherActionListener scoreMatcherListener = new MatcherActionListener(true);
    scoreMatcherListener.createScorePanel(createGroupLayout(scorePanel));

    // Compare Directory
    JPanel comparePanel = new JPanel();
    comparePanel.setBorder(BorderFactory.createTitledBorder("Compare Directory"));
    MatcherActionListener compareMatcherListener = new MatcherActionListener(false);
    compareMatcherListener.createComparePanel(createGroupLayout(comparePanel));

    JPanel matcherPanel = new JPanel();
    matcherPanel.setName("MatcherPanel");
    GroupLayout layout = createGroupLayout(matcherPanel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(scorePanel).addComponent(comparePanel));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(scorePanel).addComponent(comparePanel));
    return new JScrollPane(matcherPanel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
        ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
  }

  /**
   * Sets a groupLayout for the given jPanel
   * 
   * @param panel
   *          the JPanel to set the groupLayout
   * 
   * @return the groupLayout for the given jPanel
   */
  static GroupLayout createGroupLayout(JPanel panel) {
    GroupLayout layout = new GroupLayout(panel);
    panel.setLayout(layout);
    layout.setAutoCreateGaps(true);
    layout.setAutoCreateContainerGaps(true);
    return layout;
  }

  /**
   * Creates a click-able link to a file in the file system
   * 
   * @param filename
   *          the string file name to link to
   * @param linkButton
   *          the JButton to hold the link
   * @param split
   *          the boolean indicating whether the file is a splitter output file
   */
  static void createLink(final String filename, final JButton linkButton, boolean split) {

    Vector<JButton> linkVector = new Vector<JButton>();
    linkVector.add(linkButton);
    if (linkButton.getText().contains("Existing Dictionary") && (new File(filename)).exists()) {
      if (existingDictionaryListener.getLoadDictionaryLink().getActionCommand().length() < 1) {
        existingDictionaryListener.getLoadDictionaryLink().setActionCommand(
            createFingerprinterListener.getDictionaryLink().getActionCommand());
        linkVector.add(existingDictionaryListener.getLoadDictionaryLink());
      }
      if (createFingerprinterListener.getDictionaryLink().getActionCommand().length() < 1) {
        createFingerprinterListener.getDictionaryLink().setActionCommand(
            existingDictionaryListener.getLoadDictionaryLink().getActionCommand());
        linkVector.add(createFingerprinterListener.getDictionaryLink());
      }
    }

    for (final JButton link : linkVector) {
      link.setText("<HTML><FONT color=\"#000099\"><U>" + filename + "</U></FONT></HTML>");
      link.setToolTipText("Open " + filename);
      if (link.getActionListeners().length > 0 && !split) {
        link.removeActionListener(link.getActionListeners()[0]);
      }

      link.addActionListener(new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
          if (Desktop.isDesktopSupported()) {
            try {
              Desktop.getDesktop().open(new File(filename));
            } catch (IOException exception) {
              link.setText("");
            }
          } else {
            link.setText("");
          }
        }
      });
    }
  }

  /**
   * Sets up a JButton to look like click-able link
   * 
   * @param link
   *          the link JButton
   */
  static void linkSetup(JButton link) {
    link.setBorderPainted(false);
    link.setHorizontalAlignment(SwingConstants.LEADING);
    link.setOpaque(false);
    link.setBackground(Color.WHITE);
  }

  /**
   * Provides a panel that contains the provided components and (optionally bold) label
   * 
   * @param components
   *          list of components to display (order matters)
   * @param labelText
   *          string name of label for the panel
   * @param bold
   *          boolean value indicating whether to make the label text bold
   * 
   * @return the JPanel containing the label and components
   */
  static JPanel createSubPanel(List<JComponent> components, String labelText, boolean bold) {
    JLabel label = new JLabel(labelText);
    if (bold) {
      label.setFont(Global.boldUnifontFont);
    }
    JPanel outputPanel = new JPanel();
    GroupLayout layout = GUI.createGroupLayout(outputPanel);
    ParallelGroup parallelGroup = layout.createParallelGroup();
    SequentialGroup sequentialGroup = layout.createSequentialGroup();
    parallelGroup.addComponent(label);
    sequentialGroup.addComponent(label);
    for (JComponent component : components) {
      parallelGroup.addComponent(component);
      sequentialGroup.addComponent(component);
      sequentialGroup.addPreferredGap(label, component, LayoutStyle.ComponentPlacement.INDENT);
    }
    layout.setHorizontalGroup(parallelGroup);
    layout.setVerticalGroup(sequentialGroup);
    return outputPanel;
  }

  /**
   * Provides a panel that allows a user to load a file or directory
   * 
   * @param panelLabel
   *          string name of panel to create
   * @param fileSelectionMode
   *          int instruction on which files to display (JFileChooser.FILES_ONLY or JFileChooser.DIRECTORIES_ONLY)
   * @param panelLink
   *          JButton link to store the loaded file or directory
   * @param buttonToEnable
   *          JButton to enable once panel link is loaded; can be null
   * @param showAll
   *          boolean value indicating which files to display (Global.ALL_FILES or Global.XML_ONLY)
   * 
   * @return the JPanel containing the label, link, and bottom that allows user to load file/directory
   */
  static JPanel createLoadPanel(final String panelLabel, final int fileSelectionMode, final JButton panelLink,
      final JButton buttonToEnable, final boolean showAll) {

    JLabel label = new JLabel(panelLabel);
    label.setFont(Global.boldUnifontFont);
    final JButton directoryButton = new JButton("Select " + panelLabel);
    directoryButton.setFont((new JLabel()).getFont());
    directoryButton.addActionListener(new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent actionEvent) {
        directoryButton.setEnabled(false);
        panelLink.setText("<HTML>Loading " + panelLabel + "...</HTML>");
        JFileChooser chooser = new JFileChooser();
        if (!showAll) {
          chooser.setFileFilter(new FileNameExtensionFilter("XML Files", "xml"));
        }
        chooser.setFileSelectionMode(fileSelectionMode);
        chooser.setDialogTitle("Choose a " + panelLabel);
        int returnVal = chooser.showOpenDialog((Component) actionEvent.getSource());
        if (returnVal == JFileChooser.APPROVE_OPTION) {
          if (panelLabel.contains("Dictionary")
              && !GUI.xmlFileTypeCheck("dictionary", chooser.getSelectedFile().getAbsolutePath())) {
            panelLink.setText("");
            panelLink.setActionCommand("");
            JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), "Must load existing dictionary",
                "Error: Dictionary", JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
          } else if (panelLabel.contains("Fingerprint File")
              && !GUI.xmlFileTypeCheck("fingerprint", chooser.getSelectedFile().getAbsolutePath())) {
            panelLink.setText("");
            panelLink.setActionCommand("");
            JOptionPane.showConfirmDialog((Component) actionEvent.getSource(), "Must load existing fingerprint",
                "Error: Fingerprinter", JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
          } else {
            panelLink.setActionCommand(chooser.getSelectedFile().getAbsolutePath());
            GUI.createLink(chooser.getSelectedFile().getAbsolutePath(), panelLink, false);
            if (buttonToEnable != null) {
              buttonToEnable.setEnabled(true);
            }
          }
          if (panelLink.getName() != null && panelLink.getName().equals("matcher_firstLink")) {
            if (containsDictionary(panelLink.getActionCommand())) {
              if (buttonToEnable != null) {
                buttonToEnable.setText("Fingerprint contains dictionary");
              }
            } else {
              if (buttonToEnable != null) {
                buttonToEnable.setText("Fingerprint does NOT contain dictionary");
              }
            }
          }
        } else {
          panelLink.setText("");
          panelLink.setActionCommand("");
        }
        directoryButton.setEnabled(true);
      }
    });

    JPanel panel = new JPanel();
    GroupLayout layout = GUI.createGroupLayout(panel);
    layout.setHorizontalGroup(layout.createParallelGroup().addComponent(label).addComponent(panelLink)
        .addComponent(directoryButton));
    layout.setVerticalGroup(layout.createSequentialGroup().addComponent(label).addComponent(panelLink)
        .addComponent(directoryButton).addPreferredGap(label, panelLink, LayoutStyle.ComponentPlacement.INDENT)
        .addPreferredGap(label, directoryButton, LayoutStyle.ComponentPlacement.INDENT));
    return panel;
  }

  /**
   * Checks that the given XML filename contains a root equal to the given type
   * 
   * @param type
   *          string type to compare to the root of the XML file
   * @param filename
   *          string XML filename to check
   * 
   * @return the boolean indicating whether the XML file's root is equal to the given type
   */
  static boolean xmlFileTypeCheck(String type, String filename) {
    try {
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(new File(filename));
      return doc.getRootElement().getName().equalsIgnoreCase(type);
    } catch (JDOMException e) {
      return false;
    } catch (IOException e) {
      return false;
    }
  }

  /**
   * Checks that the given XML file contains a dictionary
   * 
   * @param filename
   *          string filename containing the dictionary XML file
   * 
   * @return the boolean indicating whether the XML file contains a dictionary
   */
  static boolean containsDictionary(String filename) {
    try {
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(new File(filename));
      Element dictionary = doc.getRootElement().getChild("dictionary");
      if (dictionary.getChildren().isEmpty() && !(new File(dictionary.getAttributeValue("name"))).exists()) {
        return false;
      }
    } catch (JDOMException e) {
      return false;
    } catch (IOException e) {
      return false;
    } catch (NullPointerException e) {
      Exception exception = new Exception("WARNING: Choosen file does not contain fingerprint");
      JOptionPane.showConfirmDialog(null, "Exception: " + exception.getMessage(), "Error: Fingerprint",
          JOptionPane.CANCEL_OPTION, JOptionPane.ERROR_MESSAGE);
    }
    return true;
  }

  /**
   * Creates and displays the GUI; should be invoked from the event dispatch thread (for thread safety)
   */
  static void createAndShowGUI() {
    JFrame frame = new JFrame("SDText");
    frame.addWindowListener(new WindowListener() {
      @Override
      public void windowOpened(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Opened");
      }

      @Override
      public void windowClosed(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Closed");
        Global.monitor.closeBuffer();
      }

      @Override
      public void windowClosing(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Closing");
        Global.monitor.closeBuffer();
      }

      @Override
      public void windowActivated(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Activated");
      }

      @Override
      public void windowDeactivated(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Deactivated");
      }

      @Override
      public void windowDeiconified(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Deiconified");
      }

      @Override
      public void windowIconified(WindowEvent windowEvent) {
        Global.monitor.writeToBuffer("GUI", "createAndShowGUI", "Window Iconified");
      }
    });
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.add(new GUI(), BorderLayout.CENTER);

    frame.pack();
    frame.setVisible(true);
  }

  /**
   * Provides a graphical user interface to run the different SDText programs
   * 
   * @param args
   *          array of string command line arguments<br>
   */
  public static void main(String[] args) {
    try {
      SwingUtilities.invokeLater(new Runnable() {
        @Override
        public void run() {
          createAndShowGUI();
        }
      });
    } catch (Exception e) {
      ExceptionHandler.handleExceptionThrow(new Exception("TEST"));
    }
  }
}
